import torch
import torch.nn as nn
import torch.nn.functional as F

def weights_init_normal(m):
    classname = m.__class__.__name__
    #print(classname)
    if classname.find("Conv3d") != -1 or classname.find("ConvTranspose3d") != -1:
        torch.nn.init.normal_(m.weight.data, 0.0, 0.02)
    elif classname.find("BatchNorm") != -1:
        torch.nn.init.normal_(m.weight.data, 1.0, 0.02)
        torch.nn.init.constant_(m.bias.data, 0.0)
    elif classname.find("Linear") != -1:
        torch.nn.init.normal_(m.weight.data, 0.0, 0.02)

class Conv33(nn.Module):
    def __init__(self, in_feat, out_feat):
        super(Conv33, self).__init__()

        # Replace the first Conv3d layer with a MobileNetV3 depthwise separable convolution
        self.conv1 = nn.Sequential(nn.Conv3d(in_feat, out_feat, kernel_size=(3, 3, 3), stride=1, padding=1, groups=in_feat),
                                   nn.BatchNorm3d(out_feat),
                                   nn.ReLU())
        # Replace the second Conv3d layer with a MobileNetV3 pointwise convolution
        self.conv2 = nn.Sequential(nn.Conv3d(out_feat, out_feat, kernel_size=(1, 1, 1)),
                                   nn.BatchNorm3d(out_feat),
                                   nn.ReLU())
    def forward(self, inputs):
        outputs = self.conv1(inputs)
        outputs = self.conv2(outputs)
        return outputs


class Conv3x3(nn.Module):
    def __init__(self, in_feat, out_feat):
        super(Conv3x3, self).__init__()
        self.conv1 = nn.Sequential(nn.Conv3d(in_feat, out_feat,
                                             kernel_size=(3, 3, 3),
                                             stride=1,
                                             padding=1),
                                   nn.BatchNorm3d(out_feat),
                                   nn.ReLU())
    def forward(self, inputs):
        outputs = self.conv1(inputs)
        # outputs = self.conv2(outputs)
        return outputs


class UpConcat(nn.Module):
    def __init__(self, in_feat, out_feat):
        super(UpConcat, self).__init__()

        self.up = nn.UpsamplingBilinear2d(scale_factor=2)

        # self.deconv = nn.ConvTranspose2d(in_feat, out_feat,
        #                                  kernel_size=3,
        #                                  stride=1,
        #                                  dilation=1)

        self.deconv = nn.ConvTranspose3d(in_feat,
                                         out_feat,
                                         kernel_size=2,
                                         stride=2)

    def forward(self, inputs, down_outputs):
        # TODO: Upsampling required after deconv?
        # outputs = self.up(inputs)
        outputs = self.deconv(inputs)
        out = torch.cat([down_outputs, outputs], 1)
        return out


class UpSample(nn.Module):
    def __init__(self, in_feat, out_feat):
        super(UpSample, self).__init__()

        self.up = nn.Upsample(scale_factor=2, mode='nearest')

        self.deconv = nn.ConvTranspose2d(in_feat,
                                         out_feat,
                                         kernel_size=2,
                                         stride=2)

    def forward(self, inputs, down_outputs):
        # TODO: Upsampling required after deconv?
        outputs = self.up(inputs)
        # outputs = self.deconv(inputs)
        out = torch.cat([outputs, down_outputs], 1)
        return out


class BAM(nn.Module):
    def __init__(self, dim):
        super(BAM, self).__init__()
        self.layer_channel = nn.Sequential(nn.Linear(dim, int(dim / 8)),
                                           nn.Linear(int(dim / 8), dim))
        self.layer_spatial_1 = nn.Sequential(nn.Conv3d(dim, int(dim / 8), kernel_size=1, padding=0, bias=True),
                                             nn.Conv3d(int(dim / 8), int(dim / 8), kernel_size=(7, 7, 3), stride=1,
                                                       padding=(3, 3, 1),
                                                       bias=True))
        self.layer_spatial_2 = nn.Sequential(nn.Conv3d(dim, int(dim / 8), kernel_size=1, padding=0, bias=True),
                                             nn.Conv3d(int(dim / 8), int(dim / 8), kernel_size=(5, 5, 3),stride=1,
                                                       padding=(2, 2, 1),
                                                       bias=True))
        self.layer_spatial_3 = nn.Sequential(nn.Conv3d(dim, int(dim / 8), kernel_size=1, padding=0, bias=True),
                                             nn.Conv3d(int(dim / 8), int(dim / 8), kernel_size=(3, 3, 1), stride=1,
                                                       padding=(1, 1, 0),
                                                       bias=True))
        self.layer_spatial = nn.Sequential(
            nn.Conv3d(int(dim / 8) * 3, int(dim / 8), kernel_size=1, padding=0, bias=True),
            nn.Conv3d(int(dim / 8), 1, kernel_size=1, padding=0, bias=True))

    def forward(self, x):
        wh = [x.shape[2], x.shape[3], x.shape[4]]
        x1 = F.avg_pool3d(x, wh)
        x1 = x1.view(x1.shape[0], -1)
        x1 = self.layer_channel(x1)
        x1 = x1.view(x1.shape[0], -1, 1, 1, 1).expand(x.shape[0], x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        x2_1 = self.layer_spatial_1(x1)
        x2_2 = self.layer_spatial_2(x1)
        x2_3 = self.layer_spatial_3(x1)
        x2 = torch.cat([x2_1, x2_2, x2_3], dim=1)
        x2 = self.layer_spatial(x2)
        x2 = x2.expand(x.shape[0], x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        x1 = x1 + x2
        x1 = torch.sigmoid(x1)
        x = x * x1
        return x

class MSFE_net(nn.Module):
    def __init__(self,num_channels=1,num_feat=[16, 32, 64, 128]):
        super(MSFE_net, self).__init__()
        self.down1 = nn.Sequential(Conv33(num_channels, num_feat[0]))

        self.down2 = nn.Sequential(nn.MaxPool3d(kernel_size=2),
                                   Conv33(num_feat[0], num_feat[1]))

        self.down3 = nn.Sequential(nn.MaxPool3d(kernel_size=2),
                                   Conv33(num_feat[1], num_feat[2]))

        self.down4 = nn.Sequential(nn.MaxPool3d(kernel_size=2),
                                   Conv33(num_feat[2], num_feat[3]))

        self.attention_layer_channel1 = BAM(dim=16)
        self.attention_layer_channel2 = BAM(dim=32)
        self.attention_layer_channel3 = BAM(dim=64)
        self.attention_layer_channel4 = BAM(dim=128)
    def forward(self,inputs):
        # print(inputs.data.size())
        down1_feat = self.down1(inputs)
        down1_feat = self.attention_layer_channel1(down1_feat)
        # print(down1_feat.size())
        down2_feat = self.down2(down1_feat)
        down2_feat = self.attention_layer_channel2(down2_feat)
        # print(down2_feat.size())
        down3_feat = self.down3(down2_feat)
        down3_feat = self.attention_layer_channel3(down3_feat)
        # print(down3_feat.size())
        down4_feat = self.down4(down3_feat)
        down4_feat = self.attention_layer_channel4(down4_feat)
        # print(down4_feat.size())

        feature_map = [down1_feat, down2_feat, down3_feat, down4_feat]
        return feature_map,down4_feat

class EFC_net(nn.Module):
    def __init__(self):
        super(EFC_net, self).__init__()
        self.conv_block = Conv3x3(128,1)
        self.classify = nn.Sequential(
            nn.Linear(256,32),
            nn.BatchNorm1d(32),
            nn.ReLU(inplace=True),
            nn.Dropout1d(0.2),
            nn.Linear(32, 2),
            nn.Sigmoid()
            #nn.Softmax(dim=1)
        )
    def forward(self,x):
        feature_map = self.conv_block(x)
        batch_size = feature_map.shape[0]
        #print(feature_map.shape)
        feature_map = torch.flatten(feature_map, 1)#.resize((batch_size,256))
        #print(feature_map.shape)
        out = self.classify(feature_map)
        return out

class EFC_net2(nn.Module):
    def __init__(self):
        super(EFC_net2, self).__init__()
        self.conv_block = nn.Sequential(
            Conv3x3(128,256),
            nn.MaxPool3d(kernel_size=2)
        )
        self.classify = nn.Sequential(
            nn.Linear(256,32),
            nn.BatchNorm1d(32),
            nn.ReLU(inplace=True),
            nn.Dropout(0.2),
            nn.Linear(32, 2),
            nn.Sigmoid()
            #nn.Softmax(dim=1)
        )
    def forward(self,x):
        feature_map = self.conv_block(x)
        feature_map = torch.mean(feature_map,dim=[2,3,4])
        #print(feature_map.shape)
        out = self.classify(feature_map)
        return out

class PSAD_net(nn.Module):
    def __init__(self, num_classes=1):
        super(PSAD_net, self).__init__()
        num_feat = [16, 32, 64, 128]
        self.up1 = UpConcat(num_feat[3], num_feat[2])
        self.upconv1 = Conv3x3(num_feat[3], num_feat[2])

        self.up2 = UpConcat(num_feat[2], num_feat[1])
        self.upconv2 = Conv3x3(num_feat[2], num_feat[1])

        self.up3 = UpConcat(num_feat[1], num_feat[0])
        self.upconv3 = Conv3x3(num_feat[1], num_feat[0])

        self.final = nn.Sequential(nn.Conv3d(num_feat[0], num_classes, kernel_size=1))

    def forward(self,feature_map,num=0):
        down1_feat,down2_feat,down3_feat,down4_feat = feature_map

        # Sigle spad
        down1_feat = down1_feat
        down2_feat = down2_feat
        down3_feat = down3_feat
        down4_feat = down4_feat

        up1_feat = self.up1(down4_feat, down3_feat)
        up1_feat = self.upconv1(up1_feat)

        up2_feat = self.up2(up1_feat, down2_feat)
        up2_feat = self.upconv2(up2_feat)

        up3_feat = self.up3(up2_feat, down1_feat)
        up3_feat = self.upconv3(up3_feat)

        outputs = self.final(up3_feat)
        #print(outputs.size())
        return outputs

class DTAnet(nn.Module):
    def __init__(self, num_channels=1, num_classes=2):
        super(DTAnet, self).__init__()
        num_feat = [16, 32, 64, 128]
        self.msfe = MSFE_net(num_channels,num_feat)
        self.classify = EFC_net()
        #self.classify = EFC_net2()  

        self.SN_pasd = PSAD_net(num_classes)
        self.WS_pasd = PSAD_net(num_classes)

    def forward(self, inputs):

        feature_map,down4_feat = self.msfe(inputs)

        classify_out = self.classify(down4_feat)

        if self.training:
            output_sn = self.SN_pasd(feature_map)
            output_ws = self.WS_pasd(feature_map)
            return output_sn,output_ws,classify_out
        else:
            c_out = torch.argmax(classify_out,dim=1)
            if c_out == 0:
                outputs = self.SN_pasd(feature_map)
            elif c_out == 1:
                outputs = self.WS_pasd(feature_map)
            return outputs,c_out
        

if __name__ == "__main__":
    model = DTAnet(num_channels=1,num_classes=2)
    model.apply(weights_init_normal)
    input = torch.randn(2, 1, 32, 64, 64)
    out_sn,out_ws,classify_out = model(input)
    print(out_sn.shape)
    print(out_ws.shape)
    print(classify_out.shape)


import torch
import torch.nn as nn
from typing import List
from monai.networks.layers.factories import split_args
from collections.abc import Sequence
from monai.networks.layers.convutils import same_padding, calculate_out_shape
import numpy as np
from torch.nn.modules.utils import _pair, _triple
from smz.registry import MODELS


class Gate(nn.Module):
    def __init__(
            self, 
            spatial_dims: int,
            in_channels: int,
            n_experts: int,
            n_activated_experts: int,
            route_scale: float = 1.0,
            update_rate: float = 0.001,
            update_bs: int = 32,
        ):
        super().__init__()
        
        self.topk = n_activated_experts
        self.num_experts = n_experts
        self.gate = nn.Sequential(
            nn.AdaptiveAvgPool3d(1) if spatial_dims == 3 else nn.AdaptiveAvgPool2d(1),
            nn.Flatten(),
            nn.Linear(in_channels, n_experts)
        )   

        # 负载均衡偏置项（不参与梯度计算）
        self.register_buffer('bias', torch.zeros(n_experts))
        self.route_scale = route_scale

        self._current_batch_size = 0
        self.register_buffer('_current_counts', torch.zeros(n_experts))
        self.register_buffer('_history_counts', torch.zeros(n_experts))
        self.update_rate = update_rate
        self.update_bs = update_bs
        self.acc_bs = 0
        self.register_buffer('_accumulative_counts', torch.zeros(n_experts))

    def forward(self, x):
        self._current_batch_size = x.size(0)

        if self.training:
            self.update_bias()

        scores = self.gate(x).sigmoid()
        original_scores = scores
        scores = scores + self.bias
        indices = torch.topk(scores, self.topk, dim=-1)[1]
        weights = original_scores.gather(-1, indices)
        weights /= weights.sum(dim=-1, keepdim=True)
        weights *= self.route_scale

        self._current_counts = torch.bincount(
            indices.view(-1), 
            minlength=self.num_experts
        ).detach()
        
        return weights.type_as(x).reshape(self._current_batch_size, -1), indices.reshape(self._current_batch_size, -1)

    @torch.no_grad()
    def update_bias(self):
        """根据当前统计更新偏置项"""
        self._history_counts += self._current_counts
        self._accumulative_counts += self._current_counts
        self.acc_bs += self._current_batch_size

        if self.acc_bs >= self.update_bs:
            self.acc_bs = 0
            # 计算负载偏差
            load_diff = self._accumulative_counts.mean() - self._accumulative_counts
            
            self._accumulative_counts.zero_()
            # 使用符号函数更新偏置
            self.bias += self.update_rate * torch.sign(load_diff)

class Expert(nn.Module):
    def __init__(
            self, 
            spatial_dims: int,
            in_channels: int,
            out_channels: int,
            kwargs: dict = dict(type='Conv', kernel_size=3, stride=1, padding=1)
            ):
        super().__init__()

        self.out_channels = out_channels
        self.stride = kwargs.get('stride', 1)
        self.stride = _pair(self.stride) if spatial_dims == 2 else _triple(self.stride)

        self.conv = MODELS.build(cfg=dict(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=out_channels,
            **kwargs)
        )


    def forward(self, x):
        return self.conv(x)

    
@MODELS.register_module('MoeConv')
@MODELS.register_module('moeconv')
class MoeConv(nn.Module):
    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        stride: Sequence[int] | int = 1,
        kernel_size: Sequence[int] | int = 3,
        padding: Sequence[int] | int  = 1,
        dilation: Sequence[int] | int = 1,
        groups: int = 1,
        bias: bool = True,
        n_experts: int = 4,
        n_activated_experts: int = 2,
        n_shared_experts: int = 1,
        route_scale: float = 1.0,
        update_rate: float = 0.001,
        update_bs: int = 32,
        expert_kwargs: List[dict] | dict = dict(type='Conv', kernel_size=3),
    ):
        super().__init__()
        self.spatial_dims = spatial_dims
        self.out_channels = out_channels
        self.topk = n_activated_experts
        self.stride = _pair(stride) if spatial_dims == 2 else _triple(stride)
        self.kernel_size = kernel_size
        self.padding = padding

        self.gate = Gate(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            n_experts=n_experts,
            n_activated_experts=n_activated_experts,
            route_scale=route_scale,
            update_rate=update_rate,
            update_bs=update_bs,
        )

        self.experts = nn.ModuleList()
        

        self.shared_experts = nn.ModuleList(
            
            [
                MODELS.build(cfg=dict(
                    type='Conv', 
                    spatial_dims=spatial_dims,
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    dilation=dilation,
                    groups=groups,
                    bias=bias
            )) for _ in range(n_shared_experts)]
        )

        expert_kwargs = expert_kwargs if isinstance(expert_kwargs, list) else [expert_kwargs]
        length = len(expert_kwargs)
        assert n_experts % length == 0, f"n_experts {n_experts} must be divisible by the number of experts {length}"
        times = n_experts // length

        expert_kwargs = expert_kwargs*times
        for i in range(n_experts):
            kwargs = expert_kwargs[i]
            kwargs['kernel_size'] = kwargs.get('kernel_size', kernel_size)
            kwargs['padding'] = kwargs.get('padding', same_padding(kwargs['kernel_size'], kwargs.get('dilation', dilation)))
            kwargs['stride'] = kwargs.get('stride', stride)
            self.experts.append(
                Expert(
                    spatial_dims=spatial_dims,
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kwargs=kwargs
                )
            )
            
    def forward(self, x):
        B, _, *spatial_size = x.shape
        spatial_size = calculate_out_shape(spatial_size, self.kernel_size, self.stride, self.padding)

        weights, indices = self.gate(x)

        output = torch.zeros(B, self.out_channels, *spatial_size, device=x.device, dtype=x.dtype)

        counts = torch.bincount(indices.flatten(), minlength=len(self.experts)).tolist()
        
        # 遍历所有专家进行并行计算
        for expert_idx, expert in enumerate(self.experts):
            if counts[expert_idx] == 0:
                continue

            expert = self.experts[expert_idx]
            idx, top = torch.where(indices == expert_idx)
            output[idx] += expert(x[idx]) * weights[idx, top].view(-1, *[1]*(len(x.shape)-1))

        for shared_expert in self.shared_experts:
            shared_output = shared_expert(x)
            output += shared_output

        return output



if __name__ == '__main__':
    import torch
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    inputs_3d = torch.randn(2, 1, 64, 64, 64).to(device)
    moe_conv_3d = MoeConv(
        spatial_dims=3,
        in_channels=1,
        out_channels=16,
        stride=1,
        n_experts=4,
        n_activated_experts=2,
        n_shared_experts=1,
        route_scale=1.0,
        # expert_kwargs=[dict(type='ODConv', kernel_size=3), dict(type='ODConv', kernel_size=9)]
        expert_kwargs=[dict(type='Conv', kernel_size=3), dict(type='Conv', kernel_size=9)],
        update_bs=8,
    ).to(device) 
    # print(moe_conv_3d)
    # print(moe_conv_3d.gate._history_counts, moe_conv_3d.gate._current_counts)
    # print(moe_conv_3d(inputs_3d).shape)
    # print(moe_conv_3d.gate._history_counts, moe_conv_3d.gate._current_counts)
    for i in range(10000):
        x = torch.randn(2, 1, 64, 64, 64).to(device)
        y = moe_conv_3d(x)
        if i % 32 == 0:
            print(i, moe_conv_3d.gate.bias, moe_conv_3d.gate._history_counts, moe_conv_3d.gate._current_counts)

    # inputs_2d = torch.randn(2, 1, 64, 64).to(device)
    
    # moe_conv_2d = MoeConv(
    #     spatial_dims=2,
    #     in_channels=1,
    #     out_channels=16,
    #     stride=1,
    #     n_experts=4,
    #     n_activated_experts=2,
    #     n_shared_experts=1,
    #     route_scale=1.0,
    # ).to(device)
    # print(moe_conv_2d(inputs_2d).shape)


import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.modules.utils import _triple

from functools import partial
from typing import Tuple, Optional, List
import math
from timm.layers.cond_conv2d import CondConv2d
from collections.abc import Sequence

import numpy as np
from smz.registry import MODELS

def get_same_padding(x: int, kernel_size: int, stride: int, dilation: int):
    if isinstance(x, torch.Tensor):
        return torch.clamp(((x / stride).ceil() - 1) * stride + (kernel_size - 1) * dilation + 1 - x, min=0)
    else:
        return max((math.ceil(x / stride) - 1) * stride + (kernel_size - 1) * dilation + 1 - x, 0)

def pad_same(
        x,
        kernel_size: List[int],
        stride: List[int],
        dilation: List[int] = (1, 1, 1),
        value: float = 0,
):
    ih, iw, id = x.size()[-3:]
    pad_h = get_same_padding(ih, kernel_size[0], stride[0], dilation[0])
    pad_w = get_same_padding(iw, kernel_size[1], stride[1], dilation[1])
    pad_d = get_same_padding(id, kernel_size[2], stride[2], dilation[2])
    x = F.pad(x, (pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2, pad_d // 2, pad_d - pad_d // 2), value=value)
    return x

def conv3d_same(
        x,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        stride: Tuple[int, int] = (1, 1, 1),
        padding: Tuple[int, int] = (0, 0, 0),
        dilation: Tuple[int, int] = (1, 1, 1),
        groups: int = 1,
):
    x = pad_same(x, weight.shape[-3:], stride, dilation)
    return F.conv3d(x, weight, bias, stride, (0, 0, 0), dilation, groups)

# Calculate symmetric padding for a convolution
def get_padding(kernel_size: int, stride: int = 1, dilation: int = 1, **_) -> int:
    padding = ((stride - 1) + dilation * (kernel_size - 1)) // 2
    return padding

# Can SAME padding for given args be done statically?
def is_static_pad(kernel_size: int, stride: int = 1, dilation: int = 1, **_):
    return stride == 1 and (dilation * (kernel_size - 1)) % 2 == 0

def get_padding_value(padding, kernel_size, **kwargs) -> Tuple[Tuple, bool]:
    dynamic = False
    if isinstance(padding, str):
        # for any string padding, the padding will be calculated for you, one of three ways
        padding = padding.lower()
        if padding == 'same':
            # TF compatible 'SAME' padding, has a performance and GPU memory allocation impact
            if is_static_pad(kernel_size, **kwargs):
                # static case, no extra overhead
                padding = get_padding(kernel_size, **kwargs)
            else:
                # dynamic 'SAME' padding, has runtime/GPU memory overhead
                padding = 0
                dynamic = True
        elif padding == 'valid':
            # 'VALID' padding, same as padding=0
            padding = 0
        else:
            # Default to PyTorch style 'same'-ish symmetric padding
            padding = get_padding(kernel_size, **kwargs)
    return padding, dynamic


def get_condconv_initializer(initializer, num_experts, expert_shape):
    def condconv_initializer(weight):
        """CondConv initializer function."""
        num_params = np.prod(expert_shape)
        if (len(weight.shape) != 2 or weight.shape[0] != num_experts or
                weight.shape[1] != num_params):
            raise (ValueError(
                'CondConv variables must have shape [num_experts, num_params]'))
        for i in range(num_experts):
            initializer(weight[i].view(expert_shape))
    return condconv_initializer


class CondConv3d(nn.Module):
    """ Conditionally Parameterized Convolution
    Inspired by: https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/condconv/condconv_layers.py

    Grouped convolution hackery for parallel execution of the per-sample kernel filters inspired by this discussion:
    https://github.com/pytorch/pytorch/issues/17983
    """
    __constants__ = ['in_channels', 'out_channels', 'dynamic_padding']

    def __init__(self, in_channels, out_channels, kernel_size=3,
                 stride=1, padding='', dilation=1, groups=1, bias=False, num_experts=4):
        super(CondConv3d, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = _triple(kernel_size)
        self.stride = _triple(stride)
        padding_val, is_padding_dynamic = get_padding_value(
            padding, kernel_size, stride=stride, dilation=dilation)
        self.dynamic_padding = is_padding_dynamic  # if in forward to work with torchscript
        self.padding = _triple(padding_val)
        self.dilation = _triple(dilation)
        self.groups = groups
        self.num_experts = num_experts

        self.weight_shape = (self.out_channels, self.in_channels // self.groups) + self.kernel_size
        weight_num_param = 1
        for wd in self.weight_shape:
            weight_num_param *= wd
        self.weight = torch.nn.Parameter(torch.Tensor(self.num_experts, weight_num_param))

        if bias:
            self.bias_shape = (self.out_channels,)
            self.bias = torch.nn.Parameter(torch.Tensor(self.num_experts, self.out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        init_weight = get_condconv_initializer(
            partial(nn.init.kaiming_uniform_, a=math.sqrt(5)), self.num_experts, self.weight_shape)
        init_weight(self.weight)
        if self.bias is not None:
            fan_in = np.prod(self.weight_shape[1:])
            bound = 1 / math.sqrt(fan_in)
            init_bias = get_condconv_initializer(
                partial(nn.init.uniform_, a=-bound, b=bound), self.num_experts, self.bias_shape)
            init_bias(self.bias)

    def forward(self, x, routing_weights):
        B, C, H, W, D = x.shape
        weight = torch.matmul(routing_weights, self.weight)
        new_weight_shape = (B * self.out_channels, self.in_channels // self.groups) + self.kernel_size
        weight = weight.view(new_weight_shape)
        bias = None
        if self.bias is not None:
            bias = torch.matmul(routing_weights, self.bias)
            bias = bias.view(B * self.out_channels)
        # move batch elements with channels so each batch element can be efficiently convolved with separate kernel
        # reshape instead of view to work with channels_last input
        x = x.reshape(1, B * C, H, W, D)
        if self.dynamic_padding:
            out = conv3d_same(
                x, weight, bias, stride=self.stride, padding=self.padding,
                dilation=self.dilation, groups=self.groups * B)
        else:
            out = F.conv3d(
                x, weight, bias, stride=self.stride, padding=self.padding,
                dilation=self.dilation, groups=self.groups * B)
        out = out.permute([1, 0, 2, 3, 4]).view(B, self.out_channels, out.shape[-3], out.shape[-2], out.shape[-1])

        # Literal port (from TF definition)
        # x = torch.split(x, 1, 0)
        # weight = torch.split(weight, 1, 0)
        # if self.bias is not None:
        #     bias = torch.matmul(routing_weights, self.bias)
        #     bias = torch.split(bias, 1, 0)
        # else:
        #     bias = [None] * B
        # out = []
        # for xi, wi, bi in zip(x, weight, bias):
        #     wi = wi.view(*self.weight_shape)
        #     if bi is not None:
        #         bi = bi.view(*self.bias_shape)
        #     out.append(self.conv_fn(
        #         xi, wi, bi, stride=self.stride, padding=self.padding,
        #         dilation=self.dilation, groups=self.groups))
        # out = torch.cat(out, 0)
        return out
    

# https://github.com/huawei-noah/Efficient-AI-Backbones/blob/master/parameternet_pytorch/cv/parameternet.py
@MODELS.register_module()
class DynamicConv(nn.Module):
    """ Dynamic Conv layer
    """
    def __init__(
            self, 
            spatial_dims: int,
            in_channels, 
            out_channels, 
            stride: Sequence[int] | int = 1,
            kernel_size: Sequence[int] | int = 3,
            padding: str = 'same', 
            dilation: Sequence[int] | int = 1,
            groups: int = 1,
            bias: bool = False,
            num_experts: int = 4):
        super().__init__()
        # print('+++', num_experts)
        self.out_channels = out_channels
        self.stride = stride if isinstance(stride, (tuple, list, np.ndarray)) else [stride] * spatial_dims
        self.routing = nn.Linear(in_channels, num_experts)
        conv_type = CondConv2d if spatial_dims == 2 else CondConv3d
        self.cond_conv = conv_type(in_channels, out_channels, kernel_size, stride, padding, dilation,
                 groups, bias, num_experts)
        self.pool = F.adaptive_avg_pool2d if spatial_dims == 2 else F.adaptive_avg_pool3d
        
    def forward(self, x):
        pooled_inputs = self.pool(x, 1).flatten(1)  # CondConv routing
        routing_weights = torch.sigmoid(self.routing(pooled_inputs))
        x = self.cond_conv(x, routing_weights)
        return x



# ------------------------------------------------------------------------
# 3D Deformable Self-attention
# ------------------------------------------------------------------------
# Modified from Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import warnings
import math
import torch
from torch import nn
import torch.nn.functional as F
from torch.nn.init import xavier_uniform_, constant_
from ..functions.ms_deform_attn_func import ms_deform_attn_core_pytorch_3D

class MSDeformAttn(nn.Module):
    def __init__(self, d_model=256, n_levels=4, n_heads=8, n_points=4):
        """
        Multi-Scale Deformable Attention Module
        :param d_model      hidden dimension
        :param n_levels     number of feature levels
        :param n_heads      number of attention heads
        :param n_points     number of sampling points per attention head per feature level
        """
        super().__init__()
        if d_model % n_heads != 0:
            raise ValueError('d_model must be divisible by n_heads, but got {} and {}'.format(d_model, n_heads))
        _d_per_head = d_model // n_heads

        self.im2col_step = 64

        self.d_model = d_model
        self.n_levels = n_levels
        self.n_heads = n_heads
        self.n_points = n_points

        self.sampling_offsets = nn.Linear(d_model, n_heads * n_levels * n_points * 3)
        self.attention_weights = nn.Linear(d_model, n_heads * n_levels * n_points)
        self.value_proj = nn.Linear(d_model, d_model)
        self.output_proj = nn.Linear(d_model, d_model)

        self._reset_parameters()

    def _reset_parameters(self):
        constant_(self.sampling_offsets.weight.data, 0.)
        thetas = torch.arange(self.n_heads, dtype=torch.float32) * (2.0 * math.pi / self.n_heads)
        grid_init = torch.stack([thetas.cos(), thetas.sin()*thetas.cos(), thetas.sin()*thetas.sin()], -1)
        grid_init = (grid_init / grid_init.abs().max(-1, keepdim=True)[0]).view(self.n_heads, 1, 1, 3).repeat(1, self.n_levels, self.n_points, 1)
        for i in range(self.n_points):
            grid_init[:, :, i, :] *= i + 1
        with torch.no_grad():
            self.sampling_offsets.bias = nn.Parameter(grid_init.view(-1))
        constant_(self.attention_weights.weight.data, 0.)
        constant_(self.attention_weights.bias.data, 0.)
        xavier_uniform_(self.value_proj.weight.data)
        constant_(self.value_proj.bias.data, 0.)
        xavier_uniform_(self.output_proj.weight.data)
        constant_(self.output_proj.bias.data, 0.)

    def forward(self, query, reference_points, input_flatten, input_spatial_shapes, input_level_start_index, input_padding_mask=None):
        """
        :param query                       (N, Length_{query}, C)
        :param reference_points            (N, Length_{query}, n_levels, 3)
        :param input_flatten               (N, \sum_{l=0}^{L-1} D_l \cdot H_l \cdot W_l, C)
        :param input_spatial_shapes        (n_levels, 3), [(D_0, H_0, W_0), (D_1, H_1, W_1), ..., (D_{L-1}, H_{L-1}, W_{L-1})]
        :param input_level_start_index     (n_levels, ), [0, D_0*H_0*W_0, D_0*H_0*W_0+D_1*H_1*W_1, D_0*H_0*W_0+D_1*H_1*W_1+D_2*H_2*W_2, ..., D_0*H_0*W_0+D_1*H_1*W_1+...+D_{L-1}*H_{L-1}*W_{L-1}]
        :param input_padding_mask          (N, \sum_{l=0}^{L-1} D_l \cdot H_l \cdot W_l), True for padding elements, False for non-padding elements

        :return output                     (N, Length_{query}, C)
        """
        N, Len_q, _ = query.shape
        N, Len_in, _ = input_flatten.shape
        assert (input_spatial_shapes[:, 0] * input_spatial_shapes[:, 1] * input_spatial_shapes[:, 2]).sum() == Len_in

        value = self.value_proj(input_flatten)
        if input_padding_mask is not None:
            value = value.masked_fill(input_padding_mask[..., None], float(0))
        value = value.view(N, Len_in, self.n_heads, self.d_model // self.n_heads)
        sampling_offsets = self.sampling_offsets(query).view(N, Len_q, self.n_heads, self.n_levels, self.n_points, 3)
        attention_weights = self.attention_weights(query).view(N, Len_q, self.n_heads, self.n_levels * self.n_points)
        attention_weights = F.softmax(attention_weights, -1).view(N, Len_q, self.n_heads, self.n_levels, self.n_points)

        if reference_points.shape[-1] == 3:
            offset_normalizer = torch.stack([input_spatial_shapes[..., 0], input_spatial_shapes[..., 2], input_spatial_shapes[..., 1]], -1)
            sampling_locations = reference_points[:, :, None, :, None, :] \
                                 + sampling_offsets / offset_normalizer[None, None, None, :, None, :]

        output = ms_deform_attn_core_pytorch_3D(value, input_spatial_shapes, sampling_locations, attention_weights)

        output = self.output_proj(output)
        return output



from torch import nn
from collections.abc import Sequence
from torch.nn.modules.utils import _triple, _pair
import numpy as np
from torchvision.ops import DeformConv2d
try:
    from dconv_native.modules import DeformConv3d
except ImportError:
    DeformConv3d = None
from functools import partial
from monai.networks.layers.utils import get_act_layer, get_norm_layer
from monai.networks.layers.convutils import same_padding
import torch
from smz.registry import MODELS

class _DeformConvBase(nn.Module):
    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        stride: Sequence[int] | int = 1,
        kernel_size: Sequence[int] | int = 3,
        padding: Sequence[int] | int | None = None,
        dilation: Sequence[int] | int = 1,
        groups: int = 1,
        offset_groups: int = 1,
        bias: bool = True,
        modulated_act: tuple | str | None = "SIGMOID",
        deform_param_op: str = "Conv"
    ):
        super(_DeformConvBase, self).__init__()
        assert spatial_dims in [2, 3], "spatial_dims should be 2 or 3"
    
        kernel_size = _pair(kernel_size) if spatial_dims == 2 else _triple(kernel_size)

        self.num_kernel_points = np.prod(kernel_size)
        self.offset_groups = offset_groups
        self.groups = groups
        self.spatial_dims = spatial_dims
        self.stride = _pair(stride) if spatial_dims == 2 else _triple(stride)
        self.modulated_act = get_act_layer(modulated_act) if modulated_act is not None else nn.Identity()

        if padding is None:
            padding = same_padding(kernel_size, dilation)

        print('deform_param_op', deform_param_op)
        self.deform_params = MODELS.build(
            cfg=dict(
                type=deform_param_op,
                spatial_dims=spatial_dims,
                in_channels=in_channels,
                out_channels=self._deform_channels(),
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=1,
                bias=bias
            )
        )
        
        self.deform_conv_op = DeformConv2d if spatial_dims == 2 else partial(DeformConv3d, offset_groups=offset_groups)
        self.deform_conv = self.deform_conv_op(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias=bias
        )
        

    def _deform_channels(self):
        return self.spatial_dims * self.offset_groups * self.num_kernel_points

    def _get_deformation(self, input):
        return self.deform_params(input), None

    def forward(self, input):
        deformation = self._get_deformation(input)
        return self.deform_conv(input, *deformation)
    

@MODELS.register_module()
class DeformConv(_DeformConvBase):
    pass

@MODELS.register_module()
class ModulatedDeformConv(_DeformConvBase):
    def _deform_channels(self):
        self.offset_channels = self.spatial_dims * self.offset_groups * self.num_kernel_points
        self.alpha_channels = self.offset_groups * self.num_kernel_points
        return self.offset_channels + self.alpha_channels

    def _get_deformation(self, input):
        deformation = self.deform_params(input)
        offset, alpha = torch.split(
            deformation,
            [self.offset_channels, self.alpha_channels],
            dim=1
        )
        return offset, self.modulated_act(alpha)

# y(𝑝_0)=∑𝑤(𝑝_𝑛 )⋅𝑥(𝑝_0+𝑝_𝑛+Δ𝑝_𝑛 )Δ𝑚_𝑛 
# 𝑝_𝑛+Δ𝑝_𝑛是最终的形状
# generate_cross_shape_grid和generate_x_shape_grid产生的Δ𝑝_𝑛已经是最终形状，因此需要减去𝑝_𝑛
# convert_dsc_offsets产生的Δ𝑝_𝑛需要加上𝑝_𝑛才是想要的形状
def convert_dsc_offsets_3d(offsets, morph, num_kernel_points):
    offsets = offsets.float()
    center = num_kernel_points // 2

    x = offsets[:, ::3]
    y = offsets[:, 1::3]
    z = offsets[:, 2::3]
    
    # Determine axes to zero and process based on morph
    if morph == 0:
        x = 0
        process_axes = [y, z]
    elif morph == 1:
        y = 0
        process_axes = [x, z]
    elif morph == 2:
        z = 0
        process_axes = [x, y]
    else:
        raise ValueError("Invalid morph value. Must be 0, 1, 2.")

    # Process each axis
    for axis in process_axes:
        # Process front half (from start to center, inclusive)
        axis[:, :center+1] = axis[:, :center+1].flip(1).cumsum(1).flip(1)
        
        # Process back half (from center to end)
        axis[:, center:] = axis[:, center:].cumsum(1)
    
    offsets[:, ::3] = x
    offsets[:, 1::3] = y 
    offsets[:, 2::3]= z
    
    return offsets


def convert_dsc_offsets_2d(offsets, morph, num_kernel_points):
    offsets = offsets.float()
    center = num_kernel_points // 2

    x = offsets[:, ::2]
    y = offsets[:, 1::2]
    
    # Determine axes to zero and process based on morph
    if morph == 0:
        x = 0
        axis = y
    elif morph == 1:
        y = 0
        axis = x
    else:
        raise ValueError("Invalid morph value. Must be 0, 1, 2.")

    # Process each axis
    axis[:, :center+1] = axis[:, :center+1].flip(1).cumsum(1).flip(1)
    # Process back half (from center to end)
    axis[:, center:] = axis[:, center:].cumsum(1)
    
    offsets[:, ::2] = x
    offsets[:, 1::2] = y 
    
    return offsets

class _DynamicSnakeConvBase(DeformConv):
    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        stride: Sequence[int] | int = 1,
        kernel_size: Sequence[int] | int = 3,
        padding: Sequence[int] | int | None = None,
        dilation: Sequence[int] | int = 1,
        groups: int = 1,
        offset_groups: int = 1,
        bias: bool = True,
        modulated_act: tuple | str | None = "SIGMOID",
        deform_param_op: nn.Module | None = None,
        offset_norm: tuple | str | None = "batch",
        offset_act: tuple | str | None = "tanh",
        morph = None,
    ):
        self.spatial_dims = spatial_dims
        self.morph = morph
        assert self.morph is not None, "morph should be set before replacing deform conv"
        assert self.morph <= 2 and spatial_dims == 3 or self.morph < 2 and spatial_dims == 2, "morph should be 0 or 1 for 2D and 2 for 3D"

        if spatial_dims == 3:
            self.convert_offsets = convert_dsc_offsets_3d
        else:
            self.convert_offsets = convert_dsc_offsets_2d
        
        kernel_size = max(kernel_size) if isinstance(kernel_size, Sequence) else kernel_size
        if self.morph == 0:
            kernel_size = (kernel_size, 1) if spatial_dims == 2 else (kernel_size, 1, 1)
        elif self.morph == 1:
            kernel_size = (1, kernel_size) if spatial_dims == 2 else (1, kernel_size, 1)
        else:
            kernel_size = (1, 1, kernel_size)
        padding = same_padding(kernel_size, dilation)
        self.num_kernel_points = np.prod(kernel_size)
        self.offset_groups = offset_groups

        super(_DynamicSnakeConvBase, self).__init__(spatial_dims, in_channels, out_channels, stride, kernel_size, padding, dilation, groups, offset_groups, bias, modulated_act, deform_param_op)
        
        self.offset_norm = get_norm_layer(offset_norm, spatial_dims, self.spatial_dims * self.offset_groups * self.num_kernel_points)
        self.offset_act = get_act_layer(offset_act)

    def _get_deformation(self, input):
        offsets, alpha = super()._get_deformation(input)
        offsets = offsets.clone()
        return self.convert_offsets(self.offset_act(self.offset_norm(offsets)), self.morph, self.num_kernel_points), alpha   

@MODELS.register_module()    
class XDSConv(_DynamicSnakeConvBase):
    def __init__(self, *args, **kwargs):
        kwargs["morph"] = 0
        super(XDSConv, self).__init__(*args, **kwargs)

@MODELS.register_module()    
class YDSConv(_DynamicSnakeConvBase):
    def __init__(self, *args, **kwargs):
        kwargs["morph"] = 1
        super(YDSConv, self).__init__(*args, **kwargs)

@MODELS.register_module()
class ZDSConv(_DynamicSnakeConvBase):
    def __init__(self, *args, **kwargs):
        kwargs["morph"] = 2
        super(ZDSConv, self).__init__(*args, **kwargs)

class _ModulatedDynamicSnakeConvBase(_DynamicSnakeConvBase):
    def _deform_channels(self):
        self.offset_channels = self.spatial_dims * self.offset_groups * self.num_kernel_points
        self.alpha_channels = self.offset_groups * self.num_kernel_points
        return self.offset_channels + self.alpha_channels

    def _get_deformation(self, input):
        deformation = self.deform_params(input)
        offset, alpha = torch.split(
            deformation,
            [self.offset_channels, self.alpha_channels],
            dim=1
        )
        offset = offset.clone()
        return self.convert_offsets(self.offset_act(self.offset_norm(offset)), self.morph, self.num_kernel_points), self.modulated_act(alpha)

@MODELS.register_module()
class XMDSConv(_ModulatedDynamicSnakeConvBase):
    def __init__(self, *args, **kwargs):
        kwargs["morph"] = 0
        super(XMDSConv, self).__init__(*args, **kwargs)

@MODELS.register_module()
class YMDSConv(_ModulatedDynamicSnakeConvBase):
    def __init__(self, *args, **kwargs):
        kwargs["morph"] = 1
        super(YMDSConv, self).__init__(*args, **kwargs)

@MODELS.register_module()
class ZMDSConv(_ModulatedDynamicSnakeConvBase):
    def __init__(self, *args, **kwargs):
        kwargs["morph"] = 2
        super(ZMDSConv, self).__init__(*args, **kwargs)

@MODELS.register_module()
class DynamicSnakeConv(nn.Module):
    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        stride: Sequence[int] | int = 1,
        kernel_size: Sequence[int] | int = 3,
        padding: Sequence[int] | int | None = None,
        dilation: Sequence[int] | int = 1,
        groups: int = 1,
        offset_groups: int = 1,
        bias: bool = True,
        modulated_act: tuple | str | None = "SIGMOID",
        deform_param_op: nn.Module | None = 'conv',
        offset_norm: tuple | str | None = "batch",
        offset_act: tuple | str | None = "tanh",
        modulated: bool = False,
    ):
        super(DynamicSnakeConv, self).__init__()
        assert spatial_dims in [2, 3], "spatial_dims should be 2 or 3"

        self.out_channels = out_channels
        self.spatial_dims = spatial_dims
        self.stride = _pair(stride) if spatial_dims == 2 else _triple(stride)
        self.conv = MODELS.build(
            cfg=dict(
                type='Conv',
                spatial_dims=spatial_dims,
                in_channels=spatial_dims*out_channels,
                out_channels=out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                dilation=1,
                groups=1,
                bias=bias
            )
        )


        xdsc_type = XMDSConv if modulated else XDSConv
        self.xdsc = nn.Sequential(
            xdsc_type(
                spatial_dims,
                in_channels,
                out_channels,
                stride,
                kernel_size,
                padding,
                dilation,
                groups,
                offset_groups,
                bias,
                modulated_act,
                deform_param_op,
                offset_norm,
                offset_act
            ),
            get_norm_layer('batch', spatial_dims, out_channels),
            get_act_layer('relu')
        )

        ydsc_type = YMDSConv if modulated else YDSConv
        self.ydsc = nn.Sequential(
            ydsc_type(
                spatial_dims,
                in_channels,
                out_channels,
                stride,
                kernel_size,
                padding,
                dilation,
                groups,
                offset_groups,
                bias,
                modulated_act,
                deform_param_op,
                offset_norm,
                offset_act
            ),
            get_norm_layer('batch', spatial_dims, out_channels),
            get_act_layer('relu')
        )

        if spatial_dims == 3:
            zdsc_type = ZMDSConv if modulated else ZDSConv
            self.zdsc = nn.Sequential(
                zdsc_type(
                    spatial_dims,
                    in_channels,
                    out_channels,
                    stride,
                    kernel_size,
                    padding,
                    dilation,
                    groups,
                    offset_groups,
                    bias,
                    modulated_act,
                    deform_param_op,
                    offset_norm,
                    offset_act
                ),
                get_norm_layer('batch', spatial_dims, out_channels),
                get_act_layer('relu')
            )
        
        self.forward_func = self.forward_3d if spatial_dims == 3 else self.forward_2d

    def forward_3d(self, input):
        return self.conv(torch.cat([
            self.xdsc(input),
            self.ydsc(input),
            self.zdsc(input)
        ], dim=1))
    
    def forward_2d(self, input):
        return self.conv(torch.cat([
            self.xdsc(input),
            self.ydsc(input)
        ], dim=1))
    
    def forward(self, input):
        return self.forward_func(input)
            

if __name__ == "__main__":
    import torch
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    inputs = torch.randn(2, 3, 64, 64).to(device)

    conv_type = DynamicSnakeConv
    conv = conv_type(
        spatial_dims=2,
        in_channels=3,
        out_channels=16,
        stride=1,
        kernel_size=3,
        padding=1,
        dilation=1,
        groups=1,
        offset_groups=1,
        bias=True,
        modulated=True
    ).to(device)
    outputs = conv(inputs)
    print(outputs.shape)  # should be (2, 16, 64, 64)

    inputs = torch.randn(2, 3, 64, 64, 64).to(device)
    conv = conv_type(
        spatial_dims=3,
        in_channels=3,
        out_channels=16,
        stride=1,
        kernel_size=3,
        padding=1,
        dilation=1,
        groups=1,
        offset_groups=1,
        bias=True,
        modulated=True
    ).to(device)
    outputs = conv(inputs)
    print(outputs.shape)  # should be (2, 16, 64, 64, 64)





# Copyright (c) 2019, Adobe Inc. All rights reserved.
#
# This work is licensed under the Creative Commons Attribution-NonCommercial-ShareAlike
# 4.0 International Public License. To view a copy of this license, visit
# https://creativecommons.org/licenses/by-nc-sa/4.0/legalcode.

"""
自定义pytorch函数，实现一维、二维、三维张量的DWT和IDWT，未考虑边界延拓
只有当图像行列数都是偶数，且重构滤波器组低频分量长度为2时，才能精确重构，否则在边界处有误差。
"""
import torch
from torch.autograd import Function

class DWTFunction_1D(Function):
    @staticmethod
    def forward(ctx, input, matrix_Low, matrix_High):
        ctx.save_for_backward(matrix_Low, matrix_High)
        L = torch.matmul(input, matrix_Low.t())
        H = torch.matmul(input, matrix_High.t())
        return L, H
    @staticmethod
    def backward(ctx, grad_L, grad_H):
        matrix_L, matrix_H = ctx.saved_variables
        grad_input = torch.add(torch.matmul(grad_L, matrix_L), torch.matmul(grad_H, matrix_H))
        return grad_input, None, None


class IDWTFunction_1D(Function):
    @staticmethod
    def forward(ctx, input_L, input_H, matrix_L, matrix_H):
        ctx.save_for_backward(matrix_L, matrix_H)
        output = torch.add(torch.matmul(input_L, matrix_L), torch.matmul(input_H, matrix_H))
        return output
    @staticmethod
    def backward(ctx, grad_output):
        matrix_L, matrix_H = ctx.saved_variables
        grad_L = torch.matmul(grad_output, matrix_L.t())
        grad_H = torch.matmul(grad_output, matrix_H.t())
        return grad_L, grad_H, None, None


class DWTFunction_2D(Function):
    @staticmethod
    def forward(ctx, input, matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1):
        ctx.save_for_backward(matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1)
        L = torch.matmul(matrix_Low_0, input)
        H = torch.matmul(matrix_High_0, input)
        LL = torch.matmul(L, matrix_Low_1)
        LH = torch.matmul(L, matrix_High_1)
        HL = torch.matmul(H, matrix_Low_1)
        HH = torch.matmul(H, matrix_High_1)
        return LL, LH, HL, HH
    @staticmethod
    def backward(ctx, grad_LL, grad_LH, grad_HL, grad_HH):
        matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1 = ctx.saved_variables
        grad_L = torch.add(torch.matmul(grad_LL, matrix_Low_1.t()), torch.matmul(grad_LH, matrix_High_1.t()))
        grad_H = torch.add(torch.matmul(grad_HL, matrix_Low_1.t()), torch.matmul(grad_HH, matrix_High_1.t()))
        grad_input = torch.add(torch.matmul(matrix_Low_0.t(), grad_L), torch.matmul(matrix_High_0.t(), grad_H))
        return grad_input, None, None, None, None


class DWTFunction_2D_tiny(Function):
    @staticmethod
    def forward(ctx, input, matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1):
        ctx.save_for_backward(matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1)
        L = torch.matmul(matrix_Low_0, input)
        LL = torch.matmul(L, matrix_Low_1)
        return LL
    @staticmethod
    def backward(ctx, grad_LL):
        matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1 = ctx.saved_variables
        grad_L = torch.matmul(grad_LL, matrix_Low_1.t())
        grad_input = torch.matmul(matrix_Low_0.t(), grad_L)
        return grad_input, None, None, None, None


class IDWTFunction_2D(Function):
    @staticmethod
    def forward(ctx, input_LL, input_LH, input_HL, input_HH,
                matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1):
        ctx.save_for_backward(matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1)
        L = torch.add(torch.matmul(input_LL, matrix_Low_1.t()), torch.matmul(input_LH, matrix_High_1.t()))
        H = torch.add(torch.matmul(input_HL, matrix_Low_1.t()), torch.matmul(input_HH, matrix_High_1.t()))
        output = torch.add(torch.matmul(matrix_Low_0.t(), L), torch.matmul(matrix_High_0.t(), H))
        return output
    @staticmethod
    def backward(ctx, grad_output):
        matrix_Low_0, matrix_Low_1, matrix_High_0, matrix_High_1 = ctx.saved_variables
        grad_L = torch.matmul(matrix_Low_0, grad_output)
        grad_H = torch.matmul(matrix_High_0, grad_output)
        grad_LL = torch.matmul(grad_L, matrix_Low_1)
        grad_LH = torch.matmul(grad_L, matrix_High_1)
        grad_HL = torch.matmul(grad_H, matrix_Low_1)
        grad_HH = torch.matmul(grad_H, matrix_High_1)
        return grad_LL, grad_LH, grad_HL, grad_HH, None, None, None, None


class DWTFunction_3D(Function):
    @staticmethod
    def forward(ctx, input,
                matrix_Low_0, matrix_Low_1, matrix_Low_2,
                matrix_High_0, matrix_High_1, matrix_High_2):
        ctx.save_for_backward(matrix_Low_0, matrix_Low_1, matrix_Low_2,
                              matrix_High_0, matrix_High_1, matrix_High_2)
        L = torch.matmul(matrix_Low_0, input)
        H = torch.matmul(matrix_High_0, input)
        LL = torch.matmul(L, matrix_Low_1).transpose(dim0 = 2, dim1 = 3)
        LH = torch.matmul(L, matrix_High_1).transpose(dim0 = 2, dim1 = 3)
        HL = torch.matmul(H, matrix_Low_1).transpose(dim0 = 2, dim1 = 3)
        HH = torch.matmul(H, matrix_High_1).transpose(dim0 = 2, dim1 = 3)
        LLL = torch.matmul(matrix_Low_2, LL).transpose(dim0 = 2, dim1 = 3)
        LLH = torch.matmul(matrix_Low_2, LH).transpose(dim0 = 2, dim1 = 3)
        LHL = torch.matmul(matrix_Low_2, HL).transpose(dim0 = 2, dim1 = 3)
        LHH = torch.matmul(matrix_Low_2, HH).transpose(dim0 = 2, dim1 = 3)
        HLL = torch.matmul(matrix_High_2, LL).transpose(dim0 = 2, dim1 = 3)
        HLH = torch.matmul(matrix_High_2, LH).transpose(dim0 = 2, dim1 = 3)
        HHL = torch.matmul(matrix_High_2, HL).transpose(dim0 = 2, dim1 = 3)
        HHH = torch.matmul(matrix_High_2, HH).transpose(dim0 = 2, dim1 = 3)
        return LLL, LLH, LHL, LHH, HLL, HLH, HHL, HHH

    @staticmethod
    def backward(ctx, grad_LLL, grad_LLH, grad_LHL, grad_LHH,
                      grad_HLL, grad_HLH, grad_HHL, grad_HHH):
        matrix_Low_0, matrix_Low_1, matrix_Low_2, matrix_High_0, matrix_High_1, matrix_High_2 = ctx.saved_variables
        grad_LL = torch.add(torch.matmul(matrix_Low_2.t(), grad_LLL.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), grad_HLL.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        grad_LH = torch.add(torch.matmul(matrix_Low_2.t(), grad_LLH.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), grad_HLH.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        grad_HL = torch.add(torch.matmul(matrix_Low_2.t(), grad_LHL.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), grad_HHL.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        grad_HH = torch.add(torch.matmul(matrix_Low_2.t(), grad_LHH.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), grad_HHH.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        grad_L = torch.add(torch.matmul(grad_LL, matrix_Low_1.t()), torch.matmul(grad_LH, matrix_High_1.t()))
        grad_H = torch.add(torch.matmul(grad_HL, matrix_Low_1.t()), torch.matmul(grad_HH, matrix_High_1.t()))
        grad_input = torch.add(torch.matmul(matrix_Low_0.t(), grad_L), torch.matmul(matrix_High_0.t(), grad_H))
        return grad_input, None, None, None, None, None, None, None, None


class DWTFunction_3D_tiny(Function):
    @staticmethod
    def forward(ctx, input,
                matrix_Low_0, matrix_Low_1, matrix_Low_2,
                matrix_High_0, matrix_High_1, matrix_High_2):
        ctx.save_for_backward(matrix_Low_0, matrix_Low_1, matrix_Low_2,
                              matrix_High_0, matrix_High_1, matrix_High_2)
        L = torch.matmul(matrix_Low_0, input)
        LL = torch.matmul(L, matrix_Low_1).transpose(dim0 = 2, dim1 = 3)
        LLL = torch.matmul(matrix_Low_2, LL).transpose(dim0 = 2, dim1 = 3)
        return LLL

    @staticmethod
    def backward(ctx, grad_LLL):
        matrix_Low_0, matrix_Low_1, matrix_Low_2, matrix_High_0, matrix_High_1, matrix_High_2 = ctx.saved_variables
        grad_LL = torch.matmul(matrix_Low_2.t(), grad_LLL.transpose(dim0 = 2, dim1 = 3)).transpose(dim0 = 2, dim1 = 3)
        grad_L = torch.matmul(grad_LL, matrix_Low_1.t())
        grad_input = torch.matmul(matrix_Low_0.t(), grad_L)
        return grad_input, None, None, None, None, None, None, None, None


class IDWTFunction_3D(Function):
    @staticmethod
    def forward(ctx, input_LLL, input_LLH, input_LHL, input_LHH,
                     input_HLL, input_HLH, input_HHL, input_HHH,
                     matrix_Low_0, matrix_Low_1, matrix_Low_2,
                     matrix_High_0, matrix_High_1, matrix_High_2):
        ctx.save_for_backward(matrix_Low_0, matrix_Low_1, matrix_Low_2,
                              matrix_High_0, matrix_High_1, matrix_High_2)
        input_LL = torch.add(torch.matmul(matrix_Low_2.t(), input_LLL.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), input_HLL.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        input_LH = torch.add(torch.matmul(matrix_Low_2.t(), input_LLH.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), input_HLH.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        input_HL = torch.add(torch.matmul(matrix_Low_2.t(), input_LHL.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), input_HHL.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        input_HH = torch.add(torch.matmul(matrix_Low_2.t(), input_LHH.transpose(dim0 = 2, dim1 = 3)), torch.matmul(matrix_High_2.t(), input_HHH.transpose(dim0 = 2, dim1 = 3))).transpose(dim0 = 2, dim1 = 3)
        input_L = torch.add(torch.matmul(input_LL, matrix_Low_1.t()), torch.matmul(input_LH, matrix_High_1.t()))
        input_H = torch.add(torch.matmul(input_HL, matrix_Low_1.t()), torch.matmul(input_HH, matrix_High_1.t()))
        output = torch.add(torch.matmul(matrix_Low_0.t(), input_L), torch.matmul(matrix_High_0.t(), input_H))
        return output
    @staticmethod
    def backward(ctx, grad_output):
        matrix_Low_0, matrix_Low_1, matrix_Low_2, matrix_High_0, matrix_High_1, matrix_High_2 = ctx.saved_variables
        grad_L = torch.matmul(matrix_Low_0, grad_output)
        grad_H = torch.matmul(matrix_High_0, grad_output)
        grad_LL = torch.matmul(grad_L, matrix_Low_1).transpose(dim0 = 2, dim1 = 3)
        grad_LH = torch.matmul(grad_L, matrix_High_1).transpose(dim0 = 2, dim1 = 3)
        grad_HL = torch.matmul(grad_H, matrix_Low_1).transpose(dim0 = 2, dim1 = 3)
        grad_HH = torch.matmul(grad_H, matrix_High_1).transpose(dim0 = 2, dim1 = 3)
        grad_LLL = torch.matmul(matrix_Low_2, grad_LL).transpose(dim0 = 2, dim1 = 3)
        grad_LLH = torch.matmul(matrix_Low_2, grad_LH).transpose(dim0 = 2, dim1 = 3)
        grad_LHL = torch.matmul(matrix_Low_2, grad_HL).transpose(dim0 = 2, dim1 = 3)
        grad_LHH = torch.matmul(matrix_Low_2, grad_HH).transpose(dim0 = 2, dim1 = 3)
        grad_HLL = torch.matmul(matrix_High_2, grad_LL).transpose(dim0 = 2, dim1 = 3)
        grad_HLH = torch.matmul(matrix_High_2, grad_LH).transpose(dim0 = 2, dim1 = 3)
        grad_HHL = torch.matmul(matrix_High_2, grad_HL).transpose(dim0 = 2, dim1 = 3)
        grad_HHH = torch.matmul(matrix_High_2, grad_HH).transpose(dim0 = 2, dim1 = 3)
        return grad_LLL, grad_LLH, grad_LHL, grad_LHH, grad_HLL, grad_HLH, grad_HHL, grad_HHH, None, None, None, None, None, None



"""
借助于已有的卷积操作实现 DWT 和 IDWT，它们的滤波器组可以是可训练参数，初始化这些滤波器组时候使用某个给定的小波
这个方法实现的 DWT 速度快了，建议使用这个版本的；但是实现的 IDWT 速度变慢了
各个层中参数 trainable 默认设置为 False，表示不对滤波器组进行训练更新；更改为 True 则表示对滤波器组进行更新。
输入数据某个维度上的尺寸小于 kernel_size，会报错
在信号边界处无法精确重构
这些层暂时只支持2进制的标量小波。其他小波或超小波，如a进制小波、多小波、小波框架、曲波、脊波、条带波、小波框架等暂不适用
目前三维数据的边界延拓没有 'reflect'，因此即便使用 haar 小波也无法精确重建
"""

try:
    import pywt
except ImportError:
    pywt = None
import math
import torch
import torch.nn as nn
import torch.nn.functional as F


__all__ = ['DWT_1D', 'IDWT_1D', 'DWT_2D', 'IDWT_2D', 'DWT_3D', 'IDWT_3D']
Pad_Mode = ['constant', 'reflect', 'replicate', 'circular']

class DWT_1D(nn.Module):
    def __init__(self, pad_type = 'reflect', wavename = 'haar',
                 stride = 2, in_channels = 1, out_channels = None, groups = None,
                 kernel_size = None, trainable = False):
        """
        :param pad_type: 对输入数据的边界延拓方式，理论上使用对称小波如 bior2.2\bior3.3 等，同时对数据进行对称延拓，可以精确重构原数据，
                         但是脚本的实现有点问题，除非使用 haar 小波，否则无法精确重构，可能是由于 python 包 pywt 中的小波滤波器组的排列方式引起的
        :param wavename: 对滤波器初始化使用的小波，暂时只支持 2 进制的标量小波。
                         其他小波或超小波，如 a 进制小波、多小波、小波框架、曲波、脊波、条带波、小波框架等暂不适用；
                         对于 2D/3D 数据，相应的滤波器是由 1D 滤波器组进行张量相乘得到的，对应的小波称为张量小波或可分离小波，若要使用不可分离小波，则要重建脚本
                         in pywt.wavelist()
        :param stride: 采样步长，脚本设置这个值必须为2，非要设置为其他数值也是可以运行的（此时需屏蔽脚本中的 assert self.stride == 2），但是不满足小波理论；
                        若是用任意进制的小波，如3进制小波，可相应调整这个参数，但此时有更多的滤波器组，会相应分解出更多高频分量，对应的还要更新脚本内容
        :param in_channels: 输入数据的通道数
        :param out_channels: 输出数据的通道数，默认与输入数据通道数相同
        :param groups: 对通道这一维度的分组数目，这个值需要能被 in_channels 整除，
                        默认值与输入数据的通道数相同，即为 in_channels；一般的卷积操作这里默认值为 1
        :param kernel_size: 卷积核尺寸，这个参数与参数 wavename 有一定的冲突，即该参数值必须大于初始化小波滤波器长度；
                            该参数的默认值是等于初始化所用小波滤波器长度
                            若训练过程中不对滤波器组进行学习更新，即参数 trainable 设置为 False，则建议参数 kernel_size 选用默认值，因为此时除了运算量的提升，并不能带来任何增益
                            若参数 trainable 设置为 True，参数 kernel_size 应大于等于初始化所用小波的滤波器长度，此时有可能训练得到更适用于当前数据分布的滤波器组
                            个人不建议 kernel_size 的值设置的比初始化小波滤波器长度大的很多，个人建议这个超出值不要大于 3
        :param trainable: 标记是否在训练过程中更新滤波器组参数；
                          若这个参数设置为 True，且同时 groups 设置为 1 ，那么：
                                DWT层等价于多个 stride = 2 的卷积层，只是对卷积核的大小以及初始化方式不同
                                IDWT层等价于多个 stride = 2 的反卷积层操作后相加，同样卷积核的大小以及初始化方式不同

                当 out_channels 和 groups 都采用默认值时，对应的是对输入数据逐通道进行小波变换
                当 groups 取值为 1 时候，与一般的卷积操作有相似，可理解为融合数据在不同通道的相同频段内的信息
                与一般的卷积层一样，理论上这些层可以处理任意尺寸的数据。
                但是，如果输入数据某个维度上尺寸小于滤波器组长度的1/2，在重构过程中对数据延拓时会报错
                另外，我们建议输入数据各个维度上的尺寸是偶数值。

                其他各层需要说明的事项与此基本相同，不再说明。
        """
        super(DWT_1D, self).__init__()
        self.trainable = trainable
        self.kernel_size = kernel_size
        if not self.trainable:
            assert self.kernel_size == None, '若训练过程中不更新滤波器组，请将 kernel_size 设置为默认值 None'
        self.in_channels = in_channels
        self.out_channels = self.in_channels if out_channels == None else out_channels
        self.groups = self.in_channels if groups == None else groups
        assert isinstance(self.groups, int) and self.in_channels % self.groups == 0, '参数 groups 的应能被 in_channels 整除'
        self.stride = stride
        assert self.stride == 2, '目前版本，stride 只能等于 2'
        self.wavename = wavename
        self.pad_type = pad_type
        assert self.pad_type in Pad_Mode
        self.get_filters()
        self.initialization()

    def get_filters(self):
        wavelet = pywt.Wavelet(self.wavename)
        band_low = torch.tensor(wavelet.rec_lo)
        band_high = torch.tensor(wavelet.rec_hi)
        length_band = band_low.size()[0]
        self.kernel_size = length_band if self.kernel_size == None else self.kernel_size
        assert self.kernel_size >= length_band, '参数 kernel_size 的取值不能小于 初始化所用小波的滤波器长度'
        a = (self.kernel_size - length_band) // 2
        b = - (self.kernel_size - length_band - a)
        b = None if b == 0 else b
        self.filt_low  = torch.zeros(self.kernel_size)
        self.filt_high = torch.zeros(self.kernel_size)
        self.filt_low[a:b] = band_low
        self.filt_high[a:b] = band_high

    def initialization(self):
        self.filter_low = self.filt_low[None, None, :].repeat((self.out_channels, self.in_channels // self.groups, 1))
        self.filter_high = self.filt_high[None, None, :].repeat((self.out_channels, self.in_channels // self.groups, 1))
        if torch.cuda.is_available():
            self.filter_low = self.filter_low.cuda()
            self.filter_high = self.filter_high.cuda()
        if self.trainable:
            self.filter_low = nn.Parameter(self.filter_low)
            self.filter_high = nn.Parameter(self.filter_high)

        if self.kernel_size % 2 == 0:
            self.pad_sizes = [self.kernel_size // 2 - 1, self.kernel_size // 2 - 1]
        else:
            self.pad_sizes = [self.kernel_size // 2, self.kernel_size // 2]

    def forward(self, input):
        assert isinstance(input, torch.Tensor)
        assert len(input.size()) == 3
        assert input.size()[1] == self.in_channels
        input = F.pad(input, pad = self.pad_sizes, mode = self.pad_type)
        return F.conv1d(input, self.filter_low, stride = self.stride, groups = self.groups), \
               F.conv1d(input, self.filter_high, stride = self.stride, groups = self.groups)


class IDWT_1D(nn.Module):
    def __init__(self, pad_type = 'reflect', wavename = 'haar',
                 stride = 2, in_channels = 1, out_channels = None, groups = None,
                 kernel_size = None, trainable = False):
        """
            参照 DWT_1D 中的说明
            理论上，使用简单上采样和卷积实现的 IDWT 要比矩阵法计算量小、速度快，
            然而由于 Pytorch 中没有实现简单上采样，在实现 IDWT 只能用与 [1,0] 做反卷积 Deconvolution 来实现简单上采样
            这使得该方法比矩阵法实现 IDWT 速度慢非常多。
        """
        super(IDWT_1D, self).__init__()
        self.trainable = trainable
        self.kernel_size = kernel_size
        if not self.trainable:
            assert self.kernel_size == None, '若训练过程中不更新滤波器组，请将 kernel_size 设置为默认值 None'
        self.in_channels = in_channels
        self.out_channels = self.in_channels if out_channels == None else out_channels
        self.groups = self.in_channels if groups == None else groups
        assert isinstance(self.groups, int) and self.in_channels % self.groups == 0
        self.stride = stride
        assert self.stride == 2
        self.wavename = wavename
        self.pad_type = pad_type
        assert self.pad_type in Pad_Mode
        self.get_filters()
        self.initialization()

    def get_filters(self):
        wavelet = pywt.Wavelet(self.wavename)
        band_low = torch.tensor(wavelet.dec_lo)
        band_high = torch.tensor(wavelet.dec_hi)
        length_band = band_low.size()[0]
        self.kernel_size = length_band if self.kernel_size == None else self.kernel_size
        assert self.kernel_size >= length_band, '参数 kernel_size 的取值不能小于 初始化所用小波的滤波器长度'
        a = (self.kernel_size - length_band) // 2
        b = - (self.kernel_size - length_band - a)
        b = None if b == 0 else b
        self.filt_low = torch.zeros(self.kernel_size)
        self.filt_high = torch.zeros(self.kernel_size)
        self.filt_low[a:b] = band_low
        self.filt_high[a:b] = band_high

    def initialization(self):
        up_filter = torch.zeros((self.stride))
        up_filter[0] = 1.0
        up_filter = up_filter[None, None, :].repeat((self.in_channels, 1, 1))
        self.register_buffer('up_filter', up_filter)
        self.filter_low = self.filt_low[None, None, :].repeat((self.out_channels, self.in_channels // self.groups, 1))
        self.filter_high = self.filt_high[None, None, :].repeat((self.out_channels, self.in_channels // self.groups, 1))
        if torch.cuda.is_available():
            self.filter_low = self.filter_low.cuda()
            self.filter_high = self.filter_high.cuda()
            self.up_filter = self.up_filter.cuda()
        if self.trainable:
            self.filter_low = nn.Parameter(self.filter_low)
            self.filter_high = nn.Parameter(self.filter_high)
        if self.kernel_size % 2 == 0:
            self.pad_sizes = [self.kernel_size // 2 - 0, self.kernel_size // 2 - 1]
        else:
            self.pad_sizes = [self.kernel_size // 2 - 1, self.kernel_size // 2 + 1]

    def forward(self, L, H):
        assert len(L.size()) == len(H.size()) == 3
        assert L.size()[0] == H.size()[0]
        assert L.size()[1] == H.size()[1] == self.in_channels
        L = F.pad(F.conv_transpose1d(L, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        H = F.pad(F.conv_transpose1d(H, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        return F.conv1d(L, self.filter_low, stride = 1, groups = self.groups) + \
               F.conv1d(H, self.filter_high, stride = 1, groups = self.groups)


class DWT_2D(nn.Module):
    def __init__(self, pad_type = 'reflect', wavename = 'haar',
                 stride = 2, in_channels = 1, out_channels = None, groups = None,
                 kernel_size = None, trainable = False, normalized = False):
        """
            参照 DWT_1D 中的说明
        """
        super(DWT_2D, self).__init__()
        #print('in DWT: {}'.format(wavename))
        self.trainable = trainable
        self.kernel_size = kernel_size
        if not self.trainable:
            assert self.kernel_size == None, '若训练过程中不更新滤波器组，请将 kernel_size 设置为默认值 None'
        self.in_channels = in_channels
        self.out_channels = self.in_channels if out_channels == None else out_channels
        self.groups = self.in_channels if groups == None else groups
        assert isinstance(self.groups, int) and self.in_channels % self.groups == 0
        self.stride = stride
        assert self.stride == 2
        self.wavename = wavename
        self.pad_type = pad_type
        assert self.pad_type in Pad_Mode
        self.normalized = normalized
        self.get_filters()
        self.initialization()

    def get_filters(self):
        wavelet = pywt.Wavelet(self.wavename)
        band_low = torch.tensor(wavelet.rec_lo)
        band_high = torch.tensor(wavelet.rec_hi)
        length_band = band_low.size()[0]
        self.kernel_size = length_band if self.kernel_size == None else self.kernel_size
        assert self.kernel_size >= length_band, '参数 kernel_size 的取值不能小于 初始化所用小波的滤波器长度'
        a = (self.kernel_size - length_band) // 2
        b = - (self.kernel_size - length_band - a)
        b = None if b == 0 else b
        self.filt_low = torch.zeros(self.kernel_size)
        self.filt_high = torch.zeros(self.kernel_size)
        self.filt_low[a:b] = band_low
        self.filt_high[a:b] = band_high

    def initialization(self):
        self.filter_ll = self.filt_low[:, None] * self.filt_low[None,:]
        self.filter_lh = self.filt_low[:, None] * self.filt_high[None,:]
        self.filter_hl = self.filt_high[:, None] * self.filt_low[None,:]
        self.filter_hh = self.filt_high[:, None] * self.filt_high[None,:]
        self.filter_ll = self.filter_ll[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        self.filter_lh = self.filter_lh[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        self.filter_hl = self.filter_hl[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        self.filter_hh = self.filter_hh[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        if self.normalized:
            self.filter_ll = self.filter_ll / 2
            self.filter_lh = self.filter_lh * math.sqrt(2)
            self.filter_hl = self.filter_hl * math.sqrt(2)
            self.filter_hh = self.filter_hh * 2
        if torch.cuda.is_available():
            self.filter_ll = self.filter_ll.cuda()
            self.filter_lh = self.filter_lh.cuda()
            self.filter_hl = self.filter_hl.cuda()
            self.filter_hh = self.filter_hh.cuda()
        if self.trainable:
            self.filter_ll = nn.Parameter(self.filter_ll)
            self.filter_lh = nn.Parameter(self.filter_lh)
            self.filter_hl = nn.Parameter(self.filter_hl)
            self.filter_hh = nn.Parameter(self.filter_hh)
        if self.kernel_size % 2 == 0:
            self.pad_sizes = [self.kernel_size // 2 - 1, self.kernel_size // 2 - 1,
                              self.kernel_size // 2 - 1, self.kernel_size // 2 - 1]
        else:
            self.pad_sizes = [self.kernel_size // 2, self.kernel_size // 2,
                              self.kernel_size // 2, self.kernel_size // 2]

    def forward(self, input):
        assert isinstance(input, torch.Tensor)
        assert len(input.size()) == 4
        assert input.size()[1] == self.in_channels
        input = F.pad(input, pad = self.pad_sizes, mode = self.pad_type)
        return F.conv2d(input, self.filter_ll, stride = self.stride, groups = self.groups),\
               F.conv2d(input, self.filter_lh, stride = self.stride, groups = self.groups),\
               F.conv2d(input, self.filter_hl, stride = self.stride, groups = self.groups),\
               F.conv2d(input, self.filter_hh, stride = self.stride, groups = self.groups)


class DWT_2D_tiny(nn.Module):
    """
    只输出低频分量
    """
    def __init__(self, pad_type = 'reflect', wavename = 'haar',
                 stride = 2, in_channels = 1, out_channels = None, groups = None,
                 kernel_size = None, trainable = False, normalized = False):
        """
            参照 DWT_1D 中的说明
        """
        super(DWT_2D_tiny, self).__init__()
        self.trainable = trainable
        self.kernel_size = kernel_size
        if not self.trainable:
            assert self.kernel_size == None, '若训练过程中不更新滤波器组，请将 kernel_size 设置为默认值 None'
        self.in_channels = in_channels
        self.out_channels = self.in_channels if out_channels == None else out_channels
        self.groups = self.in_channels if groups == None else groups
        assert isinstance(self.groups, int) and self.in_channels % self.groups == 0
        self.stride = stride
        assert self.stride == 2
        self.wavename = wavename
        self.pad_type = pad_type
        assert self.pad_type in Pad_Mode
        self.normalized = normalized
        self.get_filters()
        self.initialization()

    def get_filters(self):
        wavelet = pywt.Wavelet(self.wavename)
        band_low = torch.tensor(wavelet.rec_lo)
        length_band = band_low.size()[0]
        self.kernel_size = length_band if self.kernel_size == None else self.kernel_size
        assert self.kernel_size >= length_band, '参数 kernel_size 的取值不能小于 初始化所用小波的滤波器长度'
        a = (self.kernel_size - length_band) // 2
        b = - (self.kernel_size - length_band - a)
        b = None if b == 0 else b
        self.filt_low = torch.zeros(self.kernel_size)
        self.filt_low[a:b] = band_low

    def initialization(self):
        self.filter_ll = self.filt_low[:, None] * self.filt_low[None,:]
        self.filter_ll = self.filter_ll[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        if self.normalized:
            self.filter_ll = self.filter_ll / 2
        if torch.cuda.is_available():
            self.filter_ll = self.filter_ll.cuda()
        if self.trainable:
            self.filter_ll = nn.Parameter(self.filter_ll)
        if self.kernel_size % 2 == 0:
            self.pad_sizes = [self.kernel_size // 2 - 1, self.kernel_size // 2 - 1,
                              self.kernel_size // 2 - 1, self.kernel_size // 2 - 1]
        else:
            self.pad_sizes = [self.kernel_size // 2, self.kernel_size // 2,
                              self.kernel_size // 2, self.kernel_size // 2]

    def forward(self, input):
        assert isinstance(input, torch.Tensor)
        assert len(input.size()) == 4
        assert input.size()[1] == self.in_channels
        input = F.pad(input, pad = self.pad_sizes, mode = self.pad_type)
        return F.conv2d(input, self.filter_ll, stride = self.stride, groups = self.groups)


class IDWT_2D(nn.Module):
    def __init__(self, pad_type = 'reflect', wavename = 'haar',
                 stride = 2, in_channels = 1, out_channels = None, groups = None,
                 kernel_size = None, trainable = False):
        """
            参照 DWT_1D 中的说明
            理论上，使用简单上采样和卷积实现的 IDWT 要比矩阵法计算量小、速度快，
            然而由于 Pytorch 中没有实现简单上采样，在实现 IDWT 只能用与 [[1,0],[0,0]] 做反卷积 Deconvolution 来实现简单上采样
            这使得该方法比矩阵法实现 IDWT 速度慢非常多。
            目前，在论文 https://arxiv.org/abs/2005.14461 中，构建 WaveSNet 事实上仍使用 DWT/IDWT 的矩阵法实现
        """
        super(IDWT_2D, self).__init__()
        self.trainable = trainable
        self.kernel_size = kernel_size
        if not self.trainable:
            assert self.kernel_size == None, '若训练过程中不更新滤波器组，请将 kernel_size 设置为默认值 None'
        self.in_channels = in_channels
        self.out_channels = self.in_channels if out_channels == None else out_channels
        self.groups = self.in_channels if groups == None else groups
        assert isinstance(self.groups, int) and self.in_channels % self.groups == 0
        self.stride = stride
        assert self.stride == 2
        self.wavename = wavename
        self.pad_type = pad_type
        assert self.pad_type in Pad_Mode
        self.get_filters()
        self.initialization()

    def get_filters(self):
        wavelet = pywt.Wavelet(self.wavename)
        band_low = torch.tensor(wavelet.dec_lo)
        band_high = torch.tensor(wavelet.dec_hi)
        length_band = band_low.size()[0]
        self.kernel_size = length_band if self.kernel_size == None else self.kernel_size
        assert self.kernel_size >= length_band, '参数 kernel_size 的取值不能小于 初始化所用小波的滤波器长度'
        a = (self.kernel_size - length_band) // 2
        b = - (self.kernel_size - length_band - a)
        b = None if b == 0 else b
        self.filt_low = torch.zeros(self.kernel_size)
        self.filt_high = torch.zeros(self.kernel_size)
        self.filt_low[a:b] = band_low
        self.filt_high[a:b] = band_high

    def initialization(self):
        self.filter_ll = self.filt_low[:, None] * self.filt_low[None,:]
        self.filter_lh = self.filt_low[:, None] * self.filt_high[None,:]
        self.filter_hl = self.filt_high[:, None] * self.filt_low[None,:]
        self.filter_hh = self.filt_high[:, None] * self.filt_high[None,:]
        self.filter_ll = self.filter_ll[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        self.filter_lh = self.filter_lh[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        self.filter_hl = self.filter_hl[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        self.filter_hh = self.filter_hh[None, None, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1))
        up_filter = torch.zeros((self.stride))
        up_filter[0] = 1
        up_filter = up_filter[:, None] * up_filter[None,:]
        up_filter = up_filter[None, None, :, :].repeat(self.out_channels, 1, 1, 1)
        self.register_buffer('up_filter', up_filter)
        if torch.cuda.is_available():
            self.filter_ll = self.filter_ll.cuda()
            self.filter_lh = self.filter_lh.cuda()
            self.filter_hl = self.filter_hl.cuda()
            self.filter_hh = self.filter_hh.cuda()
            #self.up_filter = self.up_filter.cuda()
        if self.trainable:
            self.filter_ll = nn.Parameter(self.filter_ll)
            self.filter_lh = nn.Parameter(self.filter_lh)
            self.filter_hl = nn.Parameter(self.filter_hl)
            self.filter_hh = nn.Parameter(self.filter_hh)
        if self.kernel_size % 2 == 0:
            self.pad_sizes = [self.kernel_size // 2 - 0, self.kernel_size // 2 - 1,
                              self.kernel_size // 2 - 0, self.kernel_size // 2 - 1]
        else:
            self.pad_sizes = [self.kernel_size // 2 - 1, self.kernel_size // 2 + 1,
                              self.kernel_size // 2 - 1, self.kernel_size // 2 + 1]

    def forward(self, LL, LH, HL, HH):
        assert len(LL.size()) == len(LH.size()) == len(HL.size()) == len(HH.size()) == 4
        assert LL.size()[0] == LH.size()[0] == HL.size()[0] == HH.size()[0]
        assert LL.size()[1] == LH.size()[1] == HL.size()[1] == HH.size()[1] == self.in_channels
        LL = F.conv_transpose2d(LL, self.up_filter, stride = self.stride, groups = self.in_channels)
        LH = F.conv_transpose2d(LH, self.up_filter, stride = self.stride, groups = self.in_channels)
        HL = F.conv_transpose2d(HL, self.up_filter, stride = self.stride, groups = self.in_channels)
        HH = F.conv_transpose2d(HH, self.up_filter, stride = self.stride, groups = self.in_channels)
        LL = F.pad(LL, pad = self.pad_sizes, mode = self.pad_type)
        LH = F.pad(LH, pad = self.pad_sizes, mode = self.pad_type)
        HL = F.pad(HL, pad = self.pad_sizes, mode = self.pad_type)
        HH = F.pad(HH, pad = self.pad_sizes, mode = self.pad_type)
        return F.conv2d(LL, self.filter_ll, stride = 1, groups = self.groups) + \
               F.conv2d(LH, self.filter_lh, stride = 1, groups = self.groups) + \
               F.conv2d(HL, self.filter_hl, stride = 1, groups = self.groups) + \
               F.conv2d(HH, self.filter_hh, stride = 1, groups = self.groups)


class DWT_3D(nn.Module):
    def __init__(self, pad_type = 'replicate', wavename = 'haar',
                 stride = 2, in_channels = 1, out_channels = None, groups=None,
                 kernel_size = None, trainable = False):
        """
            参照 DWT_1D 中的说明
        """
        super(DWT_3D, self).__init__()
        self.trainable = trainable
        self.kernel_size = kernel_size
        if not self.trainable:
            assert self.kernel_size == None, '若训练过程中不更新滤波器组，请将 kernel_size 设置为默认值 None'
        self.in_channels = in_channels
        self.out_channels = self.in_channels if out_channels == None else out_channels
        self.groups = self.in_channels if groups == None else groups
        assert isinstance(self.groups, int) and self.in_channels % self.groups == 0
        self.stride = stride
        assert self.stride == 2
        self.wavename = wavename
        self.pad_type = pad_type
        assert self.pad_type in Pad_Mode
        self.get_filters()
        self.initialization()

    def get_filters(self):
        wavelet = pywt.Wavelet(self.wavename)
        band_low = torch.tensor(wavelet.rec_lo)
        band_high = torch.tensor(wavelet.rec_hi)
        length_band = band_low.size()[0]
        self.kernel_size = length_band if self.kernel_size == None else self.kernel_size
        assert self.kernel_size >= length_band, '参数 kernel_size 的取值不能小于 初始化所用小波的滤波器长度'
        a = (self.kernel_size - length_band) // 2
        b = - (self.kernel_size - length_band - a)
        b = None if b == 0 else b
        self.filt_low = torch.zeros(self.kernel_size)
        self.filt_high = torch.zeros(self.kernel_size)
        self.filt_low[a:b] = band_low
        self.filt_high[a:b] = band_high

    def initialization(self):
        self.filter_lll = self.filt_low[:, None, None] * self.filt_low[None, :, None] * self.filt_low[None, None, :]
        self.filter_llh = self.filt_low[:, None, None] * self.filt_low[None, :, None] * self.filt_high[None, None, :]
        self.filter_lhl = self.filt_low[:, None, None] * self.filt_high[None, :, None] * self.filt_low[None, None, :]
        self.filter_lhh = self.filt_low[:, None, None] * self.filt_high[None, :, None] * self.filt_high[None, None, :]
        self.filter_hll = self.filt_high[:, None, None] * self.filt_low[None, :, None] * self.filt_low[None, None, :]
        self.filter_hlh = self.filt_high[:, None, None] * self.filt_low[None, :, None] * self.filt_high[None, None, :]
        self.filter_hhl = self.filt_high[:, None, None] * self.filt_high[None, :, None] * self.filt_low[None, None, :]
        self.filter_hhh = self.filt_high[:, None, None] * self.filt_high[None, :, None] * self.filt_high[None, None, :]
        self.filter_lll = self.filter_lll[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_llh = self.filter_llh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_lhl = self.filter_lhl[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_lhh = self.filter_lhh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hll = self.filter_hll[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hlh = self.filter_hlh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hhl = self.filter_hhl[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hhh = self.filter_hhh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        if torch.cuda.is_available():
            self.filter_lll = nn.Parameter(self.filter_lll).cuda()
            self.filter_llh = nn.Parameter(self.filter_llh).cuda()
            self.filter_lhl = nn.Parameter(self.filter_lhl).cuda()
            self.filter_lhh = nn.Parameter(self.filter_lhh).cuda()
            self.filter_hll = nn.Parameter(self.filter_hll).cuda()
            self.filter_hlh = nn.Parameter(self.filter_hlh).cuda()
            self.filter_hhl = nn.Parameter(self.filter_hhl).cuda()
            self.filter_hhh = nn.Parameter(self.filter_hhh).cuda()
        if self.trainable:
            self.filter_lll = nn.Parameter(self.filter_lll)
            self.filter_llh = nn.Parameter(self.filter_llh)
            self.filter_lhl = nn.Parameter(self.filter_lhl)
            self.filter_lhh = nn.Parameter(self.filter_lhh)
            self.filter_hll = nn.Parameter(self.filter_hll)
            self.filter_hlh = nn.Parameter(self.filter_hlh)
            self.filter_hhl = nn.Parameter(self.filter_hhl)
            self.filter_hhh = nn.Parameter(self.filter_hhh)
        if self.kernel_size % 2 == 0:
            self.pad_sizes = [self.kernel_size // 2 - 1, self.kernel_size // 2 - 1,
                              self.kernel_size // 2 - 1, self.kernel_size // 2 - 1,
                              self.kernel_size // 2 - 1, self.kernel_size // 2 - 1]
        else:
            self.pad_sizes = [self.kernel_size // 2, self.kernel_size // 2,
                              self.kernel_size // 2, self.kernel_size // 2,
                              self.kernel_size // 2, self.kernel_size // 2]

    def forward(self, input):
        assert isinstance(input, torch.Tensor)
        assert len(input.size()) == 5
        assert input.size()[1] == self.in_channels
        input = F.pad(input, pad = self.pad_sizes, mode = self.pad_type)
        return F.conv3d(input, self.filter_lll, stride = self.stride, groups = self.groups), \
               F.conv3d(input, self.filter_llh, stride = self.stride, groups = self.groups), \
               F.conv3d(input, self.filter_lhl, stride = self.stride, groups = self.groups), \
               F.conv3d(input, self.filter_lhh, stride = self.stride, groups = self.groups), \
               F.conv3d(input, self.filter_hll, stride = self.stride, groups = self.groups), \
               F.conv3d(input, self.filter_hlh, stride = self.stride, groups = self.groups), \
               F.conv3d(input, self.filter_hhl, stride = self.stride, groups = self.groups), \
               F.conv3d(input, self.filter_hhh, stride = self.stride, groups = self.groups)


class IDWT_3D(nn.Module):
    def __init__(self, pad_type = 'replicate', wavename = 'haar',
                 stride = 2, in_channels = 1, out_channels = None, groups=None,
                 kernel_size = None, trainable=False):
        """
            参照 DWT_1D 中的说明
            理论上，使用简单上采样和卷积实现的 IDWT 要比矩阵法计算量小、速度快，
            然而由于 Pytorch 中没有实现简单上采样，在实现 IDWT 只能用与 [[[1,0],[0,0]], [[0,0],[0,0]]] 做反卷积 Deconvolution 来实现简单上采样
            这使得该方法比矩阵法实现 IDWT 速度慢非常多。
        """
        super(IDWT_3D, self).__init__()
        self.trainable = trainable
        self.kernel_size = kernel_size
        if not self.trainable:
            assert self.kernel_size == None, '若训练过程中不更新滤波器组，请将 kernel_size 设置为默认值 None'
        self.in_channels = in_channels
        self.out_channels = self.in_channels if out_channels == None else out_channels
        self.groups = self.in_channels if groups == None else groups
        assert isinstance(self.groups, int) and self.in_channels % self.groups == 0
        self.stride = stride
        assert self.stride == 2
        self.wavename = wavename
        self.pad_type = pad_type
        assert self.pad_type in Pad_Mode
        self.get_filters()
        self.initialization()

    def get_filters(self):
        wavelet = pywt.Wavelet(self.wavename)
        band_low = torch.tensor(wavelet.dec_lo)
        band_high = torch.tensor(wavelet.dec_hi)
        length_band = band_low.size()[0]
        self.kernel_size = length_band if self.kernel_size == None else self.kernel_size
        assert self.kernel_size >= length_band, '参数 kernel_size 的取值不能小于 初始化所用小波的滤波器长度'
        a = (self.kernel_size - length_band) // 2
        b = - (self.kernel_size - length_band - a)
        b = None if b == 0 else b
        self.filt_low = torch.zeros(self.kernel_size)
        self.filt_high = torch.zeros(self.kernel_size)
        self.filt_low[a:b] = band_low
        self.filt_high[a:b] = band_high

    def initialization(self):
        self.filter_lll = self.filt_low[:, None, None] * self.filt_low[None, :, None] * self.filt_low[None, None, :]
        self.filter_llh = self.filt_low[:, None, None] * self.filt_low[None, :, None] * self.filt_high[None, None, :]
        self.filter_lhl = self.filt_low[:, None, None] * self.filt_high[None, :, None] * self.filt_low[None, None, :]
        self.filter_lhh = self.filt_low[:, None, None] * self.filt_high[None, :, None] * self.filt_high[None, None, :]
        self.filter_hll = self.filt_high[:, None, None] * self.filt_low[None, :, None] * self.filt_low[None, None, :]
        self.filter_hlh = self.filt_high[:, None, None] * self.filt_low[None, :, None] * self.filt_high[None, None, :]
        self.filter_hhl = self.filt_high[:, None, None] * self.filt_high[None, :, None] * self.filt_low[None, None, :]
        self.filter_hhh = self.filt_high[:, None, None] * self.filt_high[None, :, None] * self.filt_high[None, None, :]
        self.filter_lll = self.filter_lll[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_llh = self.filter_llh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_lhl = self.filter_lhl[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_lhh = self.filter_lhh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hll = self.filter_hll[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hlh = self.filter_hlh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hhl = self.filter_hhl[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        self.filter_hhh = self.filter_hhh[None, None, :, :, :].repeat((self.out_channels, self.in_channels // self.groups, 1, 1, 1))
        up_filter = torch.zeros((self.stride))
        up_filter[0] = 1
        up_filter = up_filter[:, None, None] * up_filter[None,:, None] * up_filter[None, None, :]
        up_filter = up_filter[None, None, :, :, :].repeat(self.out_channels, 1, 1, 1, 1)
        self.register_buffer('up_filter', up_filter)
        if torch.cuda.is_available():
            self.filter_lll = nn.Parameter(self.filter_lll).cuda()
            self.filter_llh = nn.Parameter(self.filter_llh).cuda()
            self.filter_lhl = nn.Parameter(self.filter_lhl).cuda()
            self.filter_lhh = nn.Parameter(self.filter_lhh).cuda()
            self.filter_hll = nn.Parameter(self.filter_hll).cuda()
            self.filter_hlh = nn.Parameter(self.filter_hlh).cuda()
            self.filter_hhl = nn.Parameter(self.filter_hhl).cuda()
            self.filter_hhh = nn.Parameter(self.filter_hhh).cuda()
            self.up_filter = self.up_filter.cuda()
        if self.trainable:
            self.filter_lll = nn.Parameter(self.filter_lll)
            self.filter_llh = nn.Parameter(self.filter_llh)
            self.filter_lhl = nn.Parameter(self.filter_lhl)
            self.filter_lhh = nn.Parameter(self.filter_lhh)
            self.filter_hll = nn.Parameter(self.filter_hll)
            self.filter_hlh = nn.Parameter(self.filter_hlh)
            self.filter_hhl = nn.Parameter(self.filter_hhl)
            self.filter_hhh = nn.Parameter(self.filter_hhh)
        if self.kernel_size % 2 == 0:
            self.pad_sizes = [self.kernel_size // 2 - 0, self.kernel_size // 2 - 1,
                              self.kernel_size // 2 - 0, self.kernel_size // 2 - 1,
                              self.kernel_size // 2 - 0, self.kernel_size // 2 - 1]
        else:
            self.pad_sizes = [self.kernel_size // 2 - 1, self.kernel_size // 2 + 1,
                              self.kernel_size // 2 - 1, self.kernel_size // 2 + 1,
                              self.kernel_size // 2 - 1, self.kernel_size // 2 + 1]

    def forward(self, LLL, LLH, LHL, LHH, HLL, HLH, HHL, HHH):
        assert len(LLL.size()) == len(LLH.size()) == len(LHL.size()) == len(LHH.size()) == len(HLL.size()) == len(HLH.size()) == len(HHL.size()) == len(HHH.size()) == 5
        assert LLL.size()[0] == LLH.size()[0] == LHL.size()[0] == LHH.size()[0] == HLL.size()[0] == HLH.size()[0] == HHL.size()[0] == HHH.size()[0]
        assert LLL.size()[1] == LLH.size()[1] == LHL.size()[1] == LHH.size()[1] == HLL.size()[1] == HLH.size()[1] == HHL.size()[1] == HHH.size()[1] == self.in_channels
        LLL = F.pad(F.conv_transpose3d(LLL, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        LLH = F.pad(F.conv_transpose3d(LLH, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        LHL = F.pad(F.conv_transpose3d(LHL, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        LHH = F.pad(F.conv_transpose3d(LHH, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        HLL = F.pad(F.conv_transpose3d(HLL, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        HLH = F.pad(F.conv_transpose3d(HLH, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        HHL = F.pad(F.conv_transpose3d(HHL, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        HHH = F.pad(F.conv_transpose3d(HHH, self.up_filter, stride = self.stride, groups = self.in_channels), pad = self.pad_sizes, mode = self.pad_type)
        return F.conv3d(LLL, self.filter_lll, stride = 1, groups = self.groups) + \
               F.conv3d(LLH, self.filter_llh, stride = 1, groups = self.groups) + \
               F.conv3d(LHL, self.filter_lhl, stride = 1, groups = self.groups) + \
               F.conv3d(LHH, self.filter_lhh, stride = 1, groups = self.groups) + \
               F.conv3d(HLL, self.filter_hll, stride = 1, groups = self.groups) + \
               F.conv3d(HLH, self.filter_hlh, stride = 1, groups = self.groups) + \
               F.conv3d(HHL, self.filter_hhl, stride = 1, groups = self.groups) + \
               F.conv3d(HHH, self.filter_hhh, stride = 1, groups = self.groups)


if __name__ == '__main__':
    """
    import numpy as np
    vector = np.array(range(3*2*8)).reshape((3,2,8))
    print(vector)
    wavename = 'haar'
    vector = torch.tensor(vector).float()
    m0 = DWT_1D(wavename = wavename, in_channels = 2, kernel_size = 12, trainable = True)
    L, H = m0(vector)
    print('L size is {}'.format(L.size()))
    print('L is {}'.format(L))
    print('H size is {}'.format(H.size()))
    print('H is {}'.format(H))
    m1 = IDWT_1D(wavename = wavename, in_channels = 2, kernel_size = 12, trainable = True)
    vector_re = m1(L, H)
    print(vector_re)
    print(vector - vector_re)
    print(torch.max(vector - vector_re), torch.min(vector - vector_re))
    """
    """
    import cv2
    import numpy as np
    from DWT_IDWT_layer import DWT_2D as DWT_2D_old
    from DWT_IDWT_layer import IDWT_2D as IDWT_2D_old

    imagename = '/home/liqiufu/Pictures/standard_test_images/lena_color_512.tif'
    image = cv2.imread(imagename)
    cv2.imshow('image', image)
    image_tensor = torch.tensor(image).float()
    image_tensor.transpose_(dim0 = -1, dim1 = 0).transpose_(dim0 = -1, dim1 = 1)
    image_tensor.unsqueeze_(dim = 0)
    print(image_tensor.size())
    wavename = 'haar'
    m = DWT_2D(wavename = wavename, in_channels = 3, kernel_size = 2, trainable = True)
    m_o = DWT_2D_old(wavename = wavename)
    LL, LH, HL, HH = m(image_tensor)
    LL_o, LH_o, HL_o, HH_o = m_o(image_tensor)
    print('LL == > {}, {}'.format(torch.max(LL - LL_o), torch.min(LL - LL_o)))
    print('LH == > {}, {}'.format(torch.max(LH - LH_o), torch.min(LH - LH_o)))
    print('HL == > {}, {}'.format(torch.max(HL - HL_o), torch.min(HL - HL_o)))
    print('HH == > {}, {}'.format(torch.max(HH - HH_o), torch.min(HH - HH_o)))
    m1 = IDWT_2D(wavename = wavename, in_channels = 3, kernel_size = 2, trainable = True)
    m1_o = IDWT_2D_old(wavename = wavename)
    image_re = m1(LL, LH, HL, HH)
    image_re_o = m1_o(LL_o, LH_o, HL_o, HH_o)
    print('image_re size is {}'.format(image_re.size()))
    print('LL size is {}'.format(LL.size()))
    # print(torch.max(torch.abs(image_tensor - image_re)), torch.min(torch.abs(image_tensor - image_re)))
    image_re.squeeze_().transpose_(dim0 = -1, dim1 = 1).transpose_(dim0 = -1, dim1 = 0)
    image_re_o.squeeze_().transpose_(dim0 = -1, dim1 = 1).transpose_(dim0 = -1, dim1 = 0)
    image_re = np.array(image_re.data)
    image_re_o = np.array(image_re_o)
    cv2.imshow('image_re', image_re / np.max(image_re))

    print(np.max(image - image_re), np.min(image - image_re))
    gap = 0
    gap = gap
    gap_ = -gap if gap != 0 else None
    print(np.max((image - image_re)[gap:gap_, gap:gap_, 2]), np.min((image - image_re)[gap:gap_, gap:gap_, :]))
    cv2.imshow('---', (image - image_re) / np.max(image - image_re))
    cv2.imshow('------', (image - image_re_o) / np.max(image - image_re_o))
    print((image - image_re)[64, 0:20, 0])
    print((image - image_re)[64, -20:, 0])
    print((image_re_o - image_re)[64, 0:20, 0])
    print((image_re_o - image_re)[64, -20:, 0])
    cv2.waitKey(0)
    """
    """
    wavename = 'haar'
    vector = torch.ones((1,1,8,8,8))
    m0 = DWT_3D(wavename = wavename, in_channels = 1, kernel_size = 4, trainable = True)
    LLL, LLH, LHL, LHH, HLL, HLH, HHL, HHH = m0(vector)
    print('LLL size is {}'.format(LLL.size()))
    m1 = IDWT_3D(wavename = wavename, in_channels = 1, kernel_size = 4, trainable = True)
    vector_re = m1(LLL, LLH, LHL, LHH, HLL, HLH, HHL, HHH)
    print(vector - vector_re)
    """




import torch.nn as nn
import torch.nn.functional as F
import torch.autograd
import numpy as np
from collections.abc import Sequence
from smz.registry import MODELS

# https://github.com/OSVAI/ODConv/blob/main/modules/odconv.py

class Attention(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, groups=1, reduction=0.0625, kernel_num=4, min_channel=16):
        super(Attention, self).__init__()
        attention_channel = max(int(in_planes * reduction), min_channel)
        self.kernel_size = kernel_size
        self.kernel_num = kernel_num
        self.temperature = 1.0

        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Conv2d(in_planes, attention_channel, 1, bias=False)
        self.bn = nn.BatchNorm2d(attention_channel)
        self.relu = nn.ReLU(inplace=True)

        self.channel_fc = nn.Conv2d(attention_channel, in_planes, 1, bias=True)
        self.func_channel = self.get_channel_attention

        if in_planes == groups and in_planes == out_planes:  # depth-wise convolution
            self.func_filter = self.skip
        else:
            self.filter_fc = nn.Conv2d(attention_channel, out_planes, 1, bias=True)
            self.func_filter = self.get_filter_attention

        if kernel_size == 1:  # point-wise convolution
            self.func_spatial = self.skip
        else:
            self.spatial_fc = nn.Conv2d(attention_channel, kernel_size * kernel_size, 1, bias=True)
            self.func_spatial = self.get_spatial_attention

        if kernel_num == 1:
            self.func_kernel = self.skip
        else:
            self.kernel_fc = nn.Conv2d(attention_channel, kernel_num, 1, bias=True)
            self.func_kernel = self.get_kernel_attention

        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            if isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def update_temperature(self, temperature):
        self.temperature = temperature

    @staticmethod
    def skip(_):
        return 1.0

    def get_channel_attention(self, x):
        channel_attention = torch.sigmoid(self.channel_fc(x).view(x.size(0), -1, 1, 1) / self.temperature)
        return channel_attention

    def get_filter_attention(self, x):
        filter_attention = torch.sigmoid(self.filter_fc(x).view(x.size(0), -1, 1, 1) / self.temperature)
        return filter_attention

    def get_spatial_attention(self, x):
        spatial_attention = self.spatial_fc(x).view(x.size(0), 1, 1, 1, self.kernel_size, self.kernel_size)
        spatial_attention = torch.sigmoid(spatial_attention / self.temperature)
        return spatial_attention

    def get_kernel_attention(self, x):
        kernel_attention = self.kernel_fc(x).view(x.size(0), -1, 1, 1, 1, 1)
        kernel_attention = F.softmax(kernel_attention / self.temperature, dim=1)
        return kernel_attention

    def forward(self, x):
        x = self.avgpool(x)
        x = self.fc(x)
        x = self.bn(x)
        x = self.relu(x)
        return self.func_channel(x), self.func_filter(x), self.func_spatial(x), self.func_kernel(x)

class ODConv2d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride=1, padding=0, dilation=1, groups=1,
                 reduction=0.0625, kernel_num=4):
        super(ODConv2d, self).__init__()
        self.in_planes = in_planes
        self.out_planes = out_planes
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.kernel_num = kernel_num
        self.attention = Attention(in_planes, out_planes, kernel_size, groups=groups,
                                   reduction=reduction, kernel_num=kernel_num)
        self.weight = nn.Parameter(torch.randn(kernel_num, out_planes, in_planes//groups, kernel_size, kernel_size),
                                   requires_grad=True)
        self._initialize_weights()

        if self.kernel_size == 1 and self.kernel_num == 1:
            self._forward_impl = self._forward_impl_pw1x
        else:
            self._forward_impl = self._forward_impl_common

    def _initialize_weights(self):
        for i in range(self.kernel_num):
            nn.init.kaiming_normal_(self.weight[i], mode='fan_out', nonlinearity='relu')

    def update_temperature(self, temperature):
        self.attention.update_temperature(temperature)

    def _forward_impl_common(self, x):
        # Multiplying channel attention (or filter attention) to weights and feature maps are equivalent,
        # while we observe that when using the latter method the models will run faster with less gpu memory cost.
        channel_attention, filter_attention, spatial_attention, kernel_attention = self.attention(x)
        batch_size, in_planes, height, width = x.size()
        print(x.shape, channel_attention.shape)
        x = x * channel_attention
        x = x.reshape(1, -1, height, width)
        aggregate_weight = spatial_attention * kernel_attention * self.weight.unsqueeze(dim=0)
        aggregate_weight = torch.sum(aggregate_weight, dim=1).view(
            [-1, self.in_planes // self.groups, self.kernel_size, self.kernel_size])
        output = F.conv2d(x, weight=aggregate_weight, bias=None, stride=self.stride, padding=self.padding,
                          dilation=self.dilation, groups=self.groups * batch_size)
        output = output.view(batch_size, self.out_planes, output.size(-2), output.size(-1))
        output = output * filter_attention
        return output

    def _forward_impl_pw1x(self, x):
        channel_attention, filter_attention, spatial_attention, kernel_attention = self.attention(x)
        x = x * channel_attention
        output = F.conv2d(x, weight=self.weight.squeeze(dim=0), bias=None, stride=self.stride, padding=self.padding,
                          dilation=self.dilation, groups=self.groups)
        output = output * filter_attention
        return output

    def forward(self, x):
        return self._forward_impl(x)


# https://github.com/i-sahajmistry/nnUNet_BraTS2023/blob/master/nnunet/ODConv.py
class Attention3d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, groups=1, reduction=0.0625, kernel_num=4, min_channel=16):
        super(Attention3d, self).__init__()
        attention_channel = max(int(in_planes * reduction), min_channel)
        self.kernel_size = kernel_size
        self.kernel_num = kernel_num
        self.temperature = 1.0

        self.avgpool = nn.AdaptiveAvgPool3d(1)
        self.fc = nn.Conv3d(in_planes, attention_channel, 1, bias=False)
        self.bn = nn.BatchNorm3d(attention_channel)
        self.relu = nn.ReLU(inplace=True)

        self.channel_fc = nn.Conv3d(attention_channel, in_planes, 1, bias=True)
        self.func_channel = self.get_channel_attention

        if in_planes == groups and in_planes == out_planes:  # depth-wise convolution
            self.func_filter = self.skip
        else:
            self.filter_fc = nn.Conv3d(attention_channel, out_planes, 1, bias=True)
            self.func_filter = self.get_filter_attention

        if kernel_size == 1:  # point-wise convolution
            self.func_spatial = self.skip
        else:
            self.spatial_fc = nn.Conv3d(attention_channel, kernel_size * kernel_size * kernel_size, 1, bias=True)
            self.func_spatial = self.get_spatial_attention

        if kernel_num == 1:
            self.func_kernel = self.skip
        else:
            self.kernel_fc = nn.Conv3d(attention_channel, kernel_num, 1, bias=True)
            self.func_kernel = self.get_kernel_attention

        self._initialize_weights()

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            if isinstance(m, nn.BatchNorm3d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def update_temperature(self, temperature):
        self.temperature = temperature

    @staticmethod
    def skip(_):
        return 1.0

    def get_channel_attention(self, x):
        channel_attention = torch.sigmoid(self.channel_fc(x).view(x.size(0), -1, 1, 1, 1) / self.temperature)
        return channel_attention

    def get_filter_attention(self, x):
        filter_attention = torch.sigmoid(self.filter_fc(x).view(x.size(0), -1, 1, 1, 1) / self.temperature)
        return filter_attention

    def get_spatial_attention(self, x):
        spatial_attention = self.spatial_fc(x).view(x.size(0), 1, 1, 1, self.kernel_size, self.kernel_size, self.kernel_size)
        spatial_attention = torch.sigmoid(spatial_attention / self.temperature)
        return spatial_attention

    def get_kernel_attention(self, x):
        kernel_attention = self.kernel_fc(x).view(x.size(0), -1, 1, 1, 1, 1, 1)
        kernel_attention = F.softmax(kernel_attention / self.temperature, dim=1)
        return kernel_attention

    def forward(self, x):
        x = self.avgpool(x)
        x = self.fc(x)
        x = self.bn(x)
        x = self.relu(x)
        return self.func_channel(x), self.func_filter(x), self.func_spatial(x), self.func_kernel(x)


class ODConv3d(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size, stride=1, padding=0, dilation=1, groups=1,
                 reduction=0.0625, kernel_num=4):
        super(ODConv3d, self).__init__()
        self.in_planes = in_planes
        self.out_planes = out_planes
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.kernel_num = kernel_num
        self.attention = Attention3d(in_planes, out_planes, kernel_size, groups=groups,
                                   reduction=reduction, kernel_num=kernel_num)
        self.weight = nn.Parameter(torch.randn(kernel_num, out_planes, in_planes//groups, kernel_size, kernel_size, kernel_size),
                                   requires_grad=True)
        self._initialize_weights()

        if self.kernel_size == 1 and self.kernel_num == 1:
            self._forward_impl = self._forward_impl_pw1x
        else:
            self._forward_impl = self._forward_impl_common

    def _initialize_weights(self):
        for i in range(self.kernel_num):
            nn.init.kaiming_normal_(self.weight[i], mode='fan_out', nonlinearity='relu')

    def update_temperature(self, temperature):
        self.attention.update_temperature(temperature)

    def _forward_impl_common(self, x):
        # Multiplying channel attention (or filter attention) to weights and feature maps are equivalent,
        # while we observe that when using the latter method the models will run faster with less gpu memory cost.
        channel_attention, filter_attention, spatial_attention, kernel_attention = self.attention(x)
        batch_size, in_planes, height, width, depth = x.size()
        x = x * channel_attention
        x = x.reshape(1, -1, height, width, depth)
        aggregate_weight = spatial_attention * kernel_attention * self.weight.unsqueeze(dim=0)
        aggregate_weight = torch.sum(aggregate_weight, dim=1).view(
            [-1, self.in_planes // self.groups, self.kernel_size, self.kernel_size, self.kernel_size])
        output = F.conv3d(x, weight=aggregate_weight, bias=None, stride=self.stride, padding=self.padding,
                          dilation=self.dilation, groups=self.groups * batch_size)
        output = output.view(batch_size, -1, output.size(-3), output.size(-2), output.size(-1))
        output = output * filter_attention
        return output

    def _forward_impl_pw1x(self, x):
        channel_attention, filter_attention, spatial_attention, kernel_attention = self.attention(x)
        x = x * channel_attention
        output = F.conv3d(x, weight=self.weight.squeeze(dim=0), bias=None, stride=self.stride, padding=self.padding,
                          dilation=self.dilation, groups=self.groups)
        output = output * filter_attention
        return output

    def forward(self, x):
        return self._forward_impl(x)

@MODELS.register_module()
class ODConv(nn.Module):
    def __init__(
            self, 
            spatial_dims: int,
            in_channels: int,
            out_channels: int,
            stride: Sequence[int] | int = 1,
            kernel_size: Sequence[int] | int = 3, 
            padding=0,
            dilation: Sequence[int] | int = 1,
            groups: int = 1,
            bias: bool = False,
            reduction=0.0625, 
            kernel_num=4):
        super(ODConv, self).__init__()

        conv = ODConv2d if spatial_dims == 2 else ODConv3d
        kernel_size = kernel_size[0]  if isinstance(kernel_size, (tuple, list, np.ndarray)) else kernel_size
        self.conv = conv(
            in_channels, out_channels, kernel_size, stride, padding, dilation, groups, reduction, kernel_num 
        )
        self.stride = stride if isinstance(stride, (tuple, list, np.ndarray)) else [stride] * spatial_dims
        self.out_channels = out_channels

    def forward(self, x):
        return self.conv(x)


if __name__ == "__main__":
    import torch
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    inputs = torch.randn(2, 3, 64, 64).to(device)

    conv_type = ODConv
    conv = conv_type(
        spatial_dims=2,
        in_channels=3,
        out_channels=16,
        stride=2,
        kernel_size=3,
        padding=1,
        dilation=1,
        groups=1,
        # bias=True
    ).to(device)
    outputs = conv(inputs)
    print(outputs.shape)  # should be (2, 16, 64, 64)

    inputs = torch.randn(2, 3, 64, 64, 64).to(device)
    conv = conv_type(
        spatial_dims=3,
        in_channels=3,
        out_channels=16,
        stride=2,
        kernel_size=3,
        padding=1,
        dilation=1,
        groups=1,
        # bias=True,
    ).to(device)
    outputs = conv(inputs)
    print(outputs.shape)  # should be (2, 16, 64, 64, 64)


from .dscnet_2d import DSCNet as DSCNet_2d, DSCNet_pro as DSCNet_2d_pro
from .dscnet_3d import DSCNet as DSCNet_3d, DSCNetSmall as DSCNet_3d_small

# This folder contains code adapted from:
# https://github.com/YaoleiQi/DSCNet
# paper
# https://arxiv.org/abs/2307.08388


import torch
from torch import nn
from smz.registry import MODELS
# This file contains code adapted from:
# https://github.com/bbyhnu/SGSNet/blob/main/model_SGSNet.py
# paper
# https://doi.org/10.1109/TMI.2021.3125777

def conv_block_3d_out(in_dim, out_dim, activation):
    return nn.Sequential(
        nn.Conv3d(in_dim, out_dim, kernel_size=1, stride=1, padding=0),
        nn.BatchNorm3d(out_dim),
        activation, )

def conv_trans_block_3d(in_dim, out_dim, activation, kernel_size=3, stride=2):
    return nn.Sequential(
        nn.ConvTranspose3d(in_dim, out_dim, kernel_size, stride, padding=1, output_padding=1),
        nn.BatchNorm3d(out_dim),
        activation, )

def max_pooling_3d():
    return nn.MaxPool3d(kernel_size=2, stride=2, padding=0)

class Heatmap_attention2(nn.Module):
    def __init__(self, in_dim, mid_dim, out_dim, activation2):
        super(Heatmap_attention2, self).__init__()
        self.W_1 = nn.Sequential(
                        nn.Conv3d(in_dim, mid_dim, kernel_size=1, stride=1, padding=0),#3 1
                        nn.BatchNorm3d(mid_dim),)
        self.relu = nn.ReLU()
        self.psi = conv_block_3d_out(mid_dim,out_dim,activation2)#sigmoid

    def forward(self, add1, add2 ,add3, out2):
        W_1 = self.W_1(add1+add2+add3)
        psi = self.relu(W_1)
        psi_out = self.psi(psi)

        return out2*psi_out# + out2

class SKConv(nn.Module):
    def __init__(self, features, WH, M, G, r, stride=1 ,L=8):
        """ Constructor
        Args:
            features: input channel dimensionality.
            WH: input spatial dimensionality, used for GAP kernel size.
            M: the number of branchs.
            G: num of convolution groups.
            r: the radio for compute d, the length of z.
            stride: stride, default 1.
            L: the minimum dim of the vector z in paper, default 32.
        """
        super(SKConv, self).__init__()
        d = max(int(features/r), L)
        self.M = M
        self.features = features
        self.convs = nn.ModuleList([])
        for i in range(M):
            self.convs.append(nn.Sequential(
                nn.Conv3d(features, features, kernel_size=(5+i*2,5+i*2,3), stride=stride, padding=(2+i,2+i,1), groups=G),#kernel_size=3+i*2   padding=1+i
                nn.BatchNorm3d(features),
                nn.ReLU()
            ))
        # self.gap = nn.AvgPool2d(int(WH/stride))
        self.fc = nn.Linear(features, d)
        self.fcs = nn.ModuleList([])
        for i in range(M):
            self.fcs.append(
                nn.Linear(d, features)
            )
        self.softmax = nn.Softmax(dim=1)
        
    def forward(self, x):   
        for i, conv in enumerate(self.convs):
            fea = conv(x).unsqueeze(dim=1)
            if i == 0:
                feas = fea
            else:
                feas = torch.cat([feas, fea], dim=1)
        fea_U = torch.sum(feas, dim=1)
        # fea_s = self.gap(fea_U).squeeze_()
        fea_s = fea_U.mean(-1).mean(-1).mean(-1)
        fea_z = self.fc(fea_s)         
        for i, fc in enumerate(self.fcs):
            vector = fc(fea_z).unsqueeze(dim=1)
            if i == 0:
                attention_vectors = vector
            else:
                attention_vectors = torch.cat([attention_vectors, vector], dim=1)
        attention_vectors = self.softmax(attention_vectors)
        attention_vectors = attention_vectors.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        fea_v = (feas * attention_vectors).sum(dim=1)
        return fea_v

#Selective Branch Module(SBM)
class SBConv(nn.Module):
    def __init__(self, features, WH=32, M=2, G=8, r=2, stride=1 ,L=8):
        """ Constructor
        Args:
            features: input channel dimensionality.
            WH: input spatial dimensionality, used for GAP kernel size.
            M: the number of branchs.
            G: num of convolution groups.
            r: the radio for compute d, the length of z.
            stride: stride, default 1.
            L: the minimum dim of the vector z in paper, default 32.
        """
        super(SBConv, self).__init__()
        d = max(int(features/r), L)
        self.M = M
        self.features = features
        # self.gap = nn.AvgPool2d(int(WH/stride))
        self.fc = nn.Linear(features, d)
        self.fcs = nn.ModuleList([])
        for i in range(M):
            self.fcs.append(
                nn.Linear(d, features)
            )
        self.softmax = nn.Softmax(dim=1)
        
    def forward(self, branch1, branch2):  
        branch1=branch1.unsqueeze(dim=1)
        branch2=branch2.unsqueeze(dim=1)
        feas = torch.cat([branch1, branch2], dim=1)
        fea_U = torch.sum(feas, dim=1)
        # fea_s = self.gap(fea_U).squeeze_()
        fea_s = fea_U.mean(-1).mean(-1).mean(-1)
        fea_z = self.fc(fea_s)         
        for i, fc in enumerate(self.fcs):
            vector = fc(fea_z).unsqueeze(dim=1)
            if i == 0:
                attention_vectors = vector
            else:
                attention_vectors = torch.cat([attention_vectors, vector], dim=1)
        attention_vectors = self.softmax(attention_vectors)
        attention_vectors = attention_vectors.unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
        fea_v = (feas * attention_vectors).sum(dim=1)
        return fea_v
    
class SKUnit(nn.Module):
    def __init__(self, in_features, out_features, WH=32, M=2, G=8, r=2, mid_features=None, stride=1, L=8):
        """ Constructor
        Args:
            in_features: input channel dimensionality.
            out_features: output channel dimensionality.
            WH: input spatial dimensionality, used for GAP kernel size.
            M: the number of branchs.
            G: num of convolution groups.
            r: the radio for compute d, the length of z.
            mid_features: the channle dim of the middle conv with stride not 1, default out_features/2.
            stride: stride.
            L: the minimum dim of the vector z in paper.
        """
        super(SKUnit, self).__init__()
        if mid_features is None:
            mid_features = int(out_features/2)
        G = mid_features
        self.feas = nn.Sequential(
            nn.Conv3d(in_features, mid_features, 1, stride=1),
            nn.BatchNorm3d(mid_features),
            SKConv(mid_features, WH, M, G, r, stride=stride, L=L),
            nn.BatchNorm3d(mid_features),
            nn.Conv3d(mid_features, out_features, 1, stride=1),
            nn.BatchNorm3d(out_features)
        )
        if in_features == out_features: # when dim not change, in could be added diectly to out
            self.shortcut = nn.Sequential()
        else: # when dim not change, in should also change dim to be added to out
            self.shortcut = nn.Sequential(
                nn.Conv3d(in_features, out_features, 1, stride=stride),
                nn.BatchNorm3d(out_features)
            )
    def forward(self, x):
        fea = self.feas(x)
        return fea + self.shortcut(x)

def skunet_block_2_3d(in_dim, out_dim, activation=False):
    return nn.Sequential(
        SKUnit(in_dim, out_dim),
        nn.ReLU(), 
        SKUnit(out_dim, out_dim),
        nn.ReLU(), )

@MODELS.register_module()
class SGSNet(nn.Module):
    def __init__(self, in_dim, out_dim, num_filters):
        super(SGSNet, self).__init__()

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.num_filters = num_filters
        activation = nn.ReLU()
        self.sigmoid = nn.Sigmoid()
        # Down sampling
        self.down_1 = skunet_block_2_3d(self.in_dim, self.num_filters, activation)
        self.pool_1 = max_pooling_3d()
        self.down_2 = skunet_block_2_3d(self.num_filters, self.num_filters * 2, activation)
        self.pool_2 = max_pooling_3d()
        self.down_3 = skunet_block_2_3d(self.num_filters * 2, self.num_filters * 4, activation)
        self.pool_3 = max_pooling_3d()
        self.down_4 = skunet_block_2_3d(self.num_filters * 4, self.num_filters * 8, activation)

        # Up sampling of SDB
        self.trans_1 = conv_trans_block_3d(self.num_filters * 8, self.num_filters * 4, activation)
        self.up_1 = skunet_block_2_3d(self.num_filters * 8, self.num_filters * 4, activation)
        self.trans_1_add = conv_trans_block_3d(self.num_filters * 4, self.num_filters, activation, kernel_size=5, stride=4)
        self.trans_2 = conv_trans_block_3d(self.num_filters * 4, self.num_filters * 2, activation)
        self.up_2 = skunet_block_2_3d(self.num_filters * 4, self.num_filters * 2, activation)
        self.trans_2_add = conv_trans_block_3d(self.num_filters * 2, self.num_filters, activation, stride=2)
        self.trans_3 = conv_trans_block_3d(self.num_filters * 2, self.num_filters * 1, activation)
        self.up_3 = skunet_block_2_3d(self.num_filters * 2, self.num_filters * 1, activation)
        self.out1 = nn.Conv3d(self.num_filters, 1, kernel_size=1, stride=1, padding=0)

        # Up sampling of MSB
        self.trans_1_seg = conv_trans_block_3d(self.num_filters * 8, self.num_filters * 4, activation)
        self.up_1_seg = skunet_block_2_3d(self.num_filters * 8, self.num_filters * 4, activation)
        self.trans_2_seg = conv_trans_block_3d(self.num_filters * 4, self.num_filters * 2, activation)
        self.up_2_seg = skunet_block_2_3d(self.num_filters * 4, self.num_filters * 2, activation)
        self.trans_3_seg = conv_trans_block_3d(self.num_filters * 2, self.num_filters * 1, activation)
        self.up_3_seg = skunet_block_2_3d(self.num_filters * 2, self.num_filters * 1, activation)

		# Structure Attention Module (SAM)
        self.heatmap_att = Heatmap_attention2(self.num_filters , self.num_filters//2, 1, self.sigmoid)
        self.SBConv_1 = SBConv(self.num_filters * 4)
        self.SBConv_2 = SBConv(self.num_filters * 2)
        self.SBConv_3 = SBConv(self.num_filters)
        self.SBConv_out = SBConv(self.num_filters)

        self.out2 = nn.Conv3d(self.num_filters, out_dim, kernel_size=1, stride=1, padding=0)
    def forward(self, x):
        # Down sampling
        down_1 = self.down_1(x) 
        pool_1 = self.pool_1(down_1)

        down_2 = self.down_2(pool_1)
        pool_2 = self.pool_2(down_2)

        down_3 = self.down_3(pool_2)
        pool_3 = self.pool_3(down_3)
        down_4 = self.down_4(pool_3)

        # Up sampling of SDB
        trans_1 = self.trans_1(down_4)
        concat_1 = torch.cat([trans_1, down_3], dim=1)
        up_1 = self.up_1(concat_1)
        trans_2 = self.trans_2(up_1) 
        concat_2 = torch.cat([trans_2, down_2], dim=1) 
        up_2 = self.up_2(concat_2) 

        trans_3 = self.trans_3(up_2)  
        concat_3 = torch.cat([trans_3, down_1], dim=1) 
        up_3 = self.up_3(concat_3)
        out1 = self.out1(up_3)

        # Up sampling of MSB
        trans_1_seg = self.trans_1_seg(down_4) 
        concat_1_seg = torch.cat([trans_1_seg, down_3], dim=1) 
        up_1_seg = self.up_1_seg(concat_1_seg)

        trans_2_seg = self.trans_2_seg(up_1_seg)
        concat_2_seg = torch.cat([trans_2_seg, down_2], dim=1) 
        up_2_seg = self.up_2_seg(concat_2_seg) 

        trans_3_seg = self.trans_3_seg(up_2_seg) 
        concat_3_seg = torch.cat([trans_3_seg, down_1], dim=1) 
        up_3_seg = self.up_3_seg(concat_3_seg) 

        # Structure Attention Module (SAM)
        SBConv_1 = self.SBConv_1(up_1, up_1_seg)
        trans_1_add = self.trans_1_add(SBConv_1)
        SBConv_2 = self.SBConv_2(up_2, up_2_seg)
        trans_2_add = self.trans_2_add(SBConv_2) 
        SBConv_3 = self.SBConv_3(up_3, up_3_seg)
        heatmap_att = self.heatmap_att(trans_1_add,trans_2_add,SBConv_3,up_3_seg)
        # Output
        out2 = self.out2(heatmap_att+up_3_seg)#(SBConv_out)  
        return out1,out2
    
# if __name__ == "__main__":
#     model = SGSNet(1, 2, 32)
#     x = torch.randn((2, 1, 128, 128, 128))
#     y1, y2 = model(x)
#     print(y1.shape)
#     print(y2.shape)